# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette import settings
from gitfourchette.forms.brandeddialog import convertToBrandedDialog
from gitfourchette.forms.signatureform import SignatureForm
from gitfourchette.forms.ui_reposettingsdialog import Ui_RepoSettingsDialog
from gitfourchette.localization import *
from gitfourchette.porcelain import Repo, GitConfigHelper
from gitfourchette.qt import *
from gitfourchette.repoprefs import RepoPrefs
from gitfourchette.toolbox import *


def validateSigPart(x: str):
    if x == "":
        return ""
    return SignatureForm.validateInput(x)


class RepoSettingsDialog(QDialog):
    def __init__(self, repo: Repo, repoPrefs: RepoPrefs, parent: QWidget):
        super().__init__(parent)
        ui = Ui_RepoSettingsDialog()
        ui.setupUi(self)
        self.ui = ui

        self.setWindowTitle(self.windowTitle().format(repo=tquoe(settings.history.getRepoNickname(repo.workdir))))

        currentNickname = settings.history.getRepoNickname(repo.workdir, strict=True)
        ui.nicknameEdit.setText(currentNickname)
        ui.nicknameEdit.setToolTip("<p>" + ui.nicknameEdit.toolTip().format(app=qAppName()))
        ui.nicknameLabel.setToolTip(ui.nicknameEdit.toolTip())

        localName, localEmail = repo.get_local_identity()
        useLocalIdentity = bool(localName or localEmail)

        globalName, globalEmail, _globalIdentityLevel = GitConfigHelper.global_identity()
        for text, edit in [(globalName, ui.nameEdit), (globalEmail, ui.emailEdit)]:
            if text:
                text += " <" + _p("RepoSettingsDialog", "from global config") + ">"
            else:
                text = "<" + _p("RepoSettingsDialog", "global git identity not configured") + ">"
            edit.setPlaceholderText(text)

        self.localNameBackup = localName
        self.localEmailBackup = localEmail

        validator = ValidatorMultiplexer(self)
        validator.setGatedWidgets(ui.buttonBox.button(QDialogButtonBox.StandardButton.Ok))
        validator.connectInput(ui.nameEdit, validateSigPart)
        validator.connectInput(ui.emailEdit, validateSigPart)
        self.validator = validator

        ui.localIdentityCheckBox.checkStateChanged.connect(self.onLocalIdentityCheckStateChanged)
        ui.localIdentityCheckBox.setChecked(not useLocalIdentity)  # hack to trigger enablement
        ui.localIdentityCheckBox.setChecked(useLocalIdentity)

        ui.keyFilePicker.setPath(repoPrefs.customKeyFile)

        validator.run()

        convertToBrandedDialog(self, subtitleText=compactPath(repo.workdir))

    def onLocalIdentityCheckStateChanged(self, state: Qt.CheckState):
        if state == Qt.CheckState.Checked:
            self.ui.nameEdit.setText(self.localNameBackup)
            self.ui.emailEdit.setText(self.localEmailBackup)
        else:
            self.localNameBackup = self.ui.nameEdit.text()
            self.localEmailBackup = self.ui.emailEdit.text()
            self.ui.nameEdit.clear()
            self.ui.emailEdit.clear()
        self.validator.run()

    def localIdentity(self) -> tuple[str, str]:
        useLocalIdentity = self.ui.localIdentityCheckBox.isChecked()
        if useLocalIdentity:
            localName = self.ui.nameEdit.text()
            localEmail = self.ui.emailEdit.text()
        else:
            localName = ""
            localEmail = ""
        return localName, localEmail



# Form implementation generated from reading ui file 'pushdialog.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_PushDialog(object):
    def setupUi(self, PushDialog):
        PushDialog.setObjectName("PushDialog")
        PushDialog.setWindowModality(Qt.WindowModality.NonModal)
        PushDialog.setEnabled(True)
        PushDialog.resize(512, 462)
        PushDialog.setSizeGripEnabled(False)
        PushDialog.setModal(True)
        self.formLayout = QFormLayout(PushDialog)
        self.formLayout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.AllNonFixedFieldsGrow)
        self.formLayout.setObjectName("formLayout")
        self.localBranchLabel = QLabel(parent=PushDialog)
        self.localBranchLabel.setObjectName("localBranchLabel")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.LabelRole, self.localBranchLabel)
        self.localBranchEdit = QComboBox(parent=PushDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.localBranchEdit.sizePolicy().hasHeightForWidth())
        self.localBranchEdit.setSizePolicy(sizePolicy)
        self.localBranchEdit.setInsertPolicy(QComboBox.InsertPolicy.NoInsert)
        self.localBranchEdit.setObjectName("localBranchEdit")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.FieldRole, self.localBranchEdit)
        self.remoteBranchLabel = QLabel(parent=PushDialog)
        self.remoteBranchLabel.setObjectName("remoteBranchLabel")
        self.formLayout.setWidget(1, QFormLayout.ItemRole.LabelRole, self.remoteBranchLabel)
        self.remoteBranchEdit = QComboBox(parent=PushDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.remoteBranchEdit.sizePolicy().hasHeightForWidth())
        self.remoteBranchEdit.setSizePolicy(sizePolicy)
        self.remoteBranchEdit.setInsertPolicy(QComboBox.InsertPolicy.NoInsert)
        self.remoteBranchEdit.setSizeAdjustPolicy(QComboBox.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon)
        self.remoteBranchEdit.setObjectName("remoteBranchEdit")
        self.formLayout.setWidget(1, QFormLayout.ItemRole.FieldRole, self.remoteBranchEdit)
        self.newRemoteBranchStackedWidget = QStackedWidget(parent=PushDialog)
        self.newRemoteBranchStackedWidget.setObjectName("newRemoteBranchStackedWidget")
        self.newRemoteBranchPage0 = QWidget()
        self.newRemoteBranchPage0.setObjectName("newRemoteBranchPage0")
        self.horizontalLayout = QHBoxLayout(self.newRemoteBranchPage0)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.remoteNameLabel = QLabel(parent=self.newRemoteBranchPage0)
        self.remoteNameLabel.setText("REMOTE/")
        self.remoteNameLabel.setObjectName("remoteNameLabel")
        self.horizontalLayout.addWidget(self.remoteNameLabel)
        self.newRemoteBranchNameEdit = QLineEdit(parent=self.newRemoteBranchPage0)
        self.newRemoteBranchNameEdit.setText("")
        self.newRemoteBranchNameEdit.setObjectName("newRemoteBranchNameEdit")
        self.horizontalLayout.addWidget(self.newRemoteBranchNameEdit)
        self.newRemoteBranchStackedWidget.addWidget(self.newRemoteBranchPage0)
        self.newRemoteBranchPage1 = QWidget()
        self.newRemoteBranchPage1.setObjectName("newRemoteBranchPage1")
        self.verticalLayout_2 = QVBoxLayout(self.newRemoteBranchPage1)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.forcePushCheckBox = QCheckBox(parent=self.newRemoteBranchPage1)
        self.forcePushCheckBox.setObjectName("forcePushCheckBox")
        self.verticalLayout_2.addWidget(self.forcePushCheckBox)
        self.newRemoteBranchStackedWidget.addWidget(self.newRemoteBranchPage1)
        self.formLayout.setWidget(2, QFormLayout.ItemRole.FieldRole, self.newRemoteBranchStackedWidget)
        self.trackCheckBox = QCheckBox(parent=PushDialog)
        self.trackCheckBox.setObjectName("trackCheckBox")
        self.formLayout.setWidget(3, QFormLayout.ItemRole.FieldRole, self.trackCheckBox)
        self.trackingLabel = QLabel(parent=PushDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.trackingLabel.sizePolicy().hasHeightForWidth())
        self.trackingLabel.setSizePolicy(sizePolicy)
        self.trackingLabel.setText("<html><head/><body><p><span style=\" font-size:small;\">Extremely long local branch name will track an absurdly long remote branch name instead of another hilariously long remote branch name.</span></p></body></html>")
        self.trackingLabel.setTextFormat(Qt.TextFormat.RichText)
        self.trackingLabel.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignTop)
        self.trackingLabel.setWordWrap(True)
        self.trackingLabel.setObjectName("trackingLabel")
        self.formLayout.setWidget(4, QFormLayout.ItemRole.FieldRole, self.trackingLabel)
        spacerItem = QSpacerItem(20, 8, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        self.formLayout.setItem(7, QFormLayout.ItemRole.LabelRole, spacerItem)
        self.groupBox = QGroupBox(parent=PushDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.groupBox.sizePolicy().hasHeightForWidth())
        self.groupBox.setSizePolicy(sizePolicy)
        self.groupBox.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignTop)
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout = QVBoxLayout(self.groupBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.statusForm = StatusForm(parent=self.groupBox)
        self.statusForm.setMinimumSize(QSize(0, 32))
        self.statusForm.setObjectName("statusForm")
        self.verticalLayout.addWidget(self.statusForm)
        self.formLayout.setWidget(8, QFormLayout.ItemRole.SpanningRole, self.groupBox)
        self.buttonBox = QDialogButtonBox(parent=PushDialog)
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.formLayout.setWidget(18, QFormLayout.ItemRole.SpanningRole, self.buttonBox)
        self.localBranchLabel.setBuddy(self.localBranchEdit)
        self.remoteBranchLabel.setBuddy(self.remoteBranchEdit)

        self.retranslateUi(PushDialog)
        self.newRemoteBranchStackedWidget.setCurrentIndex(0)
        self.buttonBox.rejected.connect(PushDialog.reject)
        PushDialog.setTabOrder(self.localBranchEdit, self.remoteBranchEdit)
        PushDialog.setTabOrder(self.remoteBranchEdit, self.forcePushCheckBox)
        PushDialog.setTabOrder(self.forcePushCheckBox, self.newRemoteBranchNameEdit)
        PushDialog.setTabOrder(self.newRemoteBranchNameEdit, self.trackCheckBox)

    def retranslateUi(self, PushDialog):
        PushDialog.setWindowTitle(_p("PushDialog", "Push branch"))
        self.localBranchLabel.setText(_p("PushDialog", "&Local branch:"))
        self.remoteBranchLabel.setText(_p("PushDialog", "Push &to:"))
        self.newRemoteBranchNameEdit.setPlaceholderText(_p("PushDialog", "Branch name on remote"))
        self.forcePushCheckBox.setText(_p("PushDialog", "&Force push"))
        self.trackCheckBox.setText(_p("PushDialog", "&Track this remote branch after pushing"))
from gitfourchette.forms.statusform import StatusForm



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations  # TODO: Remove once we can drop support for Python <= 3.13

from collections.abc import Iterable

from gitfourchette.graphview.commitinfosearch import CommitInfoSearch
from gitfourchette.graphview.commitlogmodel import CommitLogModel
from gitfourchette.localization import *
from gitfourchette.nav import NavLocator, NavFlags, NavContext
from gitfourchette.qt import *
from gitfourchette.repomodel import CommitPathspecFilter, UC_FAKEID
from gitfourchette.search.itemviewsearchprovider import ItemViewSearchProvider
from gitfourchette.tasks import QueryCommitsTouchingPath, Jump

if TYPE_CHECKING:
    from gitfourchette.graphview.graphview import GraphView


class CommitFileSearch(ItemViewSearchProvider):
    _buddy: GraphView
    _pathspecFilter: CommitPathspecFilter

    def __init__(self, parent: GraphView):
        super().__init__(parent)
        self._pathspecFilter = parent.repoModel.commitPathspecFilter
        self._pathspecFilter.resultsUpdated.connect(self._gotResults)

    def _gotResults(self):
        if self._frozen:
            return

        cpf = self._pathspecFilter
        assert cpf.isReady()
        cpf.filterOnly = self._wantFilter

        self._buddy.clFilter.updatePathspecFilter()
        self._buddy.viewport().update()

        self.setStatus(self.TermStatus.Good if cpf.matchingIds else self.TermStatus.Bad)

    # -------------------------------------------------------------------------
    # ItemViewSearchProvider implementation

    def _walkModelImpl(self, rows: Iterable[int]) -> QModelIndex:
        # We should be ready now
        assert self._pathspecFilter.isReady()
        assert self._status != self.TermStatus.Loading

        model = self.buddyModel

        for i in rows:
            index = model.index(i, 0)
            oid = model.data(index, CommitLogModel.Role.Oid)
            if oid in self._pathspecFilter.matchingIds:
                return index

        raise KeyError()

    def _jumpToIndex(self, index: QModelIndex):
        oid = self.buddyModel.data(index, CommitLogModel.Role.Oid)
        if oid == UC_FAKEID:
            delta = self._buddy.repoModel.workdirMatchesPathNeedle(self._term)
            assert delta is not None
            context = NavContext.fromGitDeltaSource(delta.source)
            locator = NavLocator(context, oid, delta.new.path)
        else:
            locator = NavLocator.inCommit(oid, self._term).withExtraFlags(NavFlags.FuzzyPath)
        Jump.invoke(self._buddy, locator)

    # -------------------------------------------------------------------------
    # SearchProvider implementation

    def longTitle(self) -> str:
        return _("Find path touched by commits")

    def shortTitle(self) -> str:
        return _("Path")

    def keyboardShortcut(self) -> str:
        return "Alt+P"

    def invalidate(self):
        self._buddy.repoWidget.taskRunner.killTaskClass(QueryCommitsTouchingPath)
        self._pathspecFilter.clear()
        self._buddy.clFilter.updatePathspecFilter()
        super().invalidate()

    def canFilter(self) -> bool:
        return True

    def notFoundMessage(self) -> str:
        message = super().notFoundMessage()
        return CommitInfoSearch.makeNotFoundMessage(message, self._buddy.repoModel)

    def _termChanged(self):
        super()._termChanged()

        # Surround with wildcards
        if self._term:
            self._term = f"*{self._term}*"

        # HACK: Always invalidate badStem when changing file searches
        # This also invalidates the pathspec filter
        self.invalidate()

    def prime(self, forwardHint: bool):
        assert self._term
        self.setStatus(self.TermStatus.Loading)
        QueryCommitsTouchingPath.invoke(self._buddy, self._term)

    def setFilterState(self, checked: bool):
        super().setFilterState(checked)
        self._pathspecFilter.filterOnly = self._wantFilter
        self._buddy.clFilter.updatePathspecFilter()



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

from gitfourchette import settings
from gitfourchette.blameview.blamemodel import BlameModel, RevList, Revision
from gitfourchette.diffview.diffdocument import LineData
from gitfourchette.gitdriver import argsIf, GitDriver, GitStatus, GitDeltaSource, GitDelta
from gitfourchette.gitdriver.parsers import parseGitBlame
from gitfourchette.localization import *
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.repomodel import UC_FAKEID
from gitfourchette.syntax import LexJobCache, LexerCache, LexJob
from gitfourchette.tasks import RepoTask, TaskPrereqs
from gitfourchette.tasks.repotask import AbortTask
from gitfourchette.toolbox import *

if TYPE_CHECKING:
    from gitfourchette.blameview.blamewindow import BlameWindow


class OpenBlame(RepoTask):
    def prereqs(self) -> TaskPrereqs:
        return TaskPrereqs.NoUnborn

    def flow(self, path: str, seed: Oid = NULL_OID, showAsync: bool = True) -> RepoTask.Flow[BlameWindow]:
        from gitfourchette.blameview.blamewindow import BlameWindow

        upperBound = yield from self._findUpperBound(path, seed)
        revList = yield from self._buildRevList(path, upperBound)

        blameModel = BlameModel(self.repoModel, revList)

        blameWindow = BlameWindow(blameModel)
        blameWindow.taskRunner.repoModel = self.repoModel
        blameWindow.exploreCommit.connect(self.rw.jump)
        blameModel.revsFile.setParent(blameWindow)

        # Die in tandem with RepoWidget
        self.rw.destroyed.connect(blameWindow.close)

        self.epilog.status = _n("{n} revision found.", "{n} revisions found.", n=len(revList))

        if showAsync:
            blameWindow.show()
            blameWindow.activateWindow()  # bring to foreground after ProcessDialog

            try:
                start = revList.revisionForCommit(seed)
            except KeyError:
                start = blameModel.currentRevision

            # Queue up another RepoTask to show the revision asynchronously
            blameWindow.showRevision(start)

        return blameWindow

    def _findUpperBound(self, path: str, seed: Oid) -> RepoTask.Flow[Oid]:
        if seed == NULL_OID:
            return seed

        # Favor current branch if it contains this commit
        if path not in self.repo.head_tree:
            useHead = False
        else:
            driver = yield from self.flowCallGit(
                "merge-base",
                "--is-ancestor",
                str(seed),
                "HEAD",
                autoFail=False)
            if driver.exitCode() not in [0, 1]:
                raise AbortTask(driver.htmlErrorText(), details=driver.formatCommandLine())
            useHead = driver.exitCode() == 0

        if useHead:
            newSeed = self.repo.head_commit_id
        else:
            # Fall back to most recent branch that contains this commit
            driver = yield from self.flowCallGit(
                "for-each-ref",
                "--count=1",
                f"--contains={seed}",
                "--format=%(objectname)")
            hashStr = driver.stdoutScrollback().strip()
            newSeed = Oid(hex=hashStr)

        if seed != newSeed:
            tree = self.repo[newSeed].peel(Tree)
            if path in tree:
                seed = newSeed

        return seed

    def _buildRevList(self, path: str, upperBound: Oid) -> RepoTask.Flow[RevList]:
        seedPath = path
        revList = RevList()

        while path:
            revision, path = yield from self._expandRevList(revList, path, upperBound)
            if revision:
                upperBound = revision.commitId

        if len(revList) == 0:
            raise AbortTask(_("File {0} has no history in the repository.", bquo(seedPath)))

        wdDelta = self.repoModel.findWorkdirDelta(seedPath)
        if wdDelta is not None:
            topRev = revList.sequence[0]
            wdRev = Revision(wdDelta.new.path, UC_FAKEID, parentIds=[topRev.commitId], status=wdDelta.status)
            revList.insert(0, wdRev)

        return revList

    def _expandRevList(
            self,
            revList: RevList,
            path: str,
            upperBound: Oid
    ) -> RepoTask.Flow[tuple[Revision | None, str]]:
        # Notes about some of the arguments:
        # --parents
        #       Enable parent rewriting so we can build a simplified graph
        #       containing just the commits that touch this file.
        # --topo-order
        #       Prevent dangling lanes in the graph. Note that this causes git
        #       to output the result in one go, making line-by-line processing
        #       futile.
        # --show-pulls
        #       Show merge commits to make the graph easier to understand.
        # We're not using "--follow" because it appears to be incompatible with
        # parent rewriting, which we need to prepare the graph. Instead, we'll
        # chain 'git log' calls if we detect a rename at the bottom of the
        # history.
        driver = yield from self.flowCallGit(
            "log",
            "--show-pulls",
            "--parents",
            "--topo-order",
            *argsIf(settings.prefs.chronologicalOrder, "--date-order"),
            *argsIf(upperBound != NULL_OID, str(upperBound)),
            "--format=%H %P",
            "--",
            path)

        scrollback = driver.stdoutScrollback()

        revision = None
        bottomPath = ""

        for line in scrollback.splitlines():
            hashes = [Oid(hex=h) for h in line.strip().split(" ")]
            commitId = hashes[0]
            parentIds = hashes[1:]
            try:
                revision = revList.revisionForCommit(commitId)
                assert not revision.parentIds, "existing revision already has parents!!!"
                revision.parentIds = parentIds
            except KeyError:
                status = GitStatus.Modified if parentIds else GitStatus.Added
                revision = Revision(path, commitId, parentIds, status=status)
                revList.push(revision)

                # Tip commit (not referred to by another commit in the trace):
                # May be a deletion or a rename
                if commitId not in revList.nonTipCommits:
                    yield from self._refineWithDelta(revision)

            revList.nonTipCommits.update(parentIds)

        # Last revision has no parents - See if it's a rename
        if revision is not None and not revision.parentIds:
            delta = yield from self._refineWithDelta(revision)
            if delta.status in [GitStatus.Renamed, GitStatus.Copied]:
                bottomPath = delta.old.path

        # TODO: If the top commit is 'R' we could expand its history upwards!
        return revision, bottomPath

    def _refineWithDelta(self, node: Revision):
        commit = self.repo.peel_commit(node.commitId)
        diffAB = commit_diff_pair(commit)
        tokens = GitDriver.buildDiffRawCommand(diffAB)
        driver = yield from self.flowCallGit(*tokens)
        deltas = driver.readDiffRawZ()
        try:
            delta = next(d for d in deltas if d.new.path == node.path)
        except StopIteration:
            delta = next(d for d in deltas if d.old.path == node.path)
            node.path = delta.new.path
        node.status = delta.status
        return delta


class OpenBlameToLine(RepoTask):
    def prereqs(self) -> TaskPrereqs:
        return TaskPrereqs.NoUnborn

    def flow(self, delta: GitDelta, lineData: LineData):
        if lineData.origin == "-":
            file = delta.old
            line = lineData.oldLineNo
        else:
            file = delta.new
            line = lineData.newLineNo

        # ----------------------------------------------------------------------
        # Find out on which commit to seed the blame

        if file.source == GitDeltaSource.Commit:
            seed = file.sourceCommit
        elif file.source == GitDeltaSource.Index and delta.source == GitDeltaSource.Dirty:
            seed = self.repo.head_commit_id
        else:
            seed = NULL_OID

        # ----------------------------------------------------------------------
        # Find out which commit introduced this line

        driver = yield from self.flowCallGit(
            "blame",
            "--porcelain",
            f"-L{line},{line}",
            *argsIf(seed != NULL_OID, str(seed)),
            "--",
            file.path)

        introducerHash, introducerLine, _dummy = next(parseGitBlame(driver.stdoutScrollback()))
        introducerCommit = Oid(hex=introducerHash)
        # Note: Git can return 0000000 if the change originated in the workdir.
        # We happen to interpret that as UC_FAKEID.

        # ----------------------------------------------------------------------
        # Perform full blame (from seed)

        blameWindow: BlameWindow = yield from self.flowSubtask(OpenBlame, path=file.path, seed=seed, showAsync=False)

        # ----------------------------------------------------------------------
        # Zero in on the commit that introduced the line

        try:
            introducerRev = blameWindow.model.revList.revisionForCommit(introducerCommit)
        except KeyError as ex:
            raise AbortTask(f"{introducerCommit} missing from revlist") from ex
        else:
            yield from self.flowSubtask(BlameRevision, introducerRev, False, blameWindow)
        finally:
            blameWindow.show()
            blameWindow.activateWindow()  # bring to foreground after ProcessDialog

        # ----------------------------------------------------------------------
        # Select the line

        block = blameWindow.textEdit.document().findBlockByNumber(introducerLine - 1)
        assert block.isValid()
        cursor = QTextCursor(block)  # cursor positioned at block start
        cursor.setPosition(block.position() + block.length() - 1, QTextCursor.MoveMode.KeepAnchor)
        blameWindow.textEdit.setTextCursor(cursor)
        blameWindow.textEdit.centerCursor()

        # Check selection accuracy
        if file.source.isWorkdir() and cursor.selectedText().rstrip() != lineData.text.rstrip():
            sorry = paragraphs(
                _("Couldn’t blame the exact line you selected, because "
                  "the file contains both staged and unstaged changes."),
                _("Try staging the entire file for more reliable results."))
            showInformation(blameWindow, _("Blame Line"), sorry)


class BlameRevision(RepoTask):
    def broadcastProcesses(self) -> bool:
        # Don't show ProcessDialog when switching to another revision
        return False

    def canKill(self, task: RepoTask) -> bool:
        # Allow interrupting a blame by switching to another revision via the
        # scrubber or nav buttons
        return isinstance(task, BlameRevision) or super().canKill(task)

    def flow(self, revision: Revision, saveAndTransposePosition: bool, blameWindow: BlameWindow | None = None):
        from gitfourchette.blameview.blamewindow import BlameWindow

        if blameWindow is None:
            blameWindow = self.parentWidget()  # type: ignore[assignment]
        assert isinstance(blameWindow, BlameWindow)

        blameModel = blameWindow.model

        # Stop lexing BEFORE changing the document!
        blameWindow.textEdit.highlighter.stopLexJobs()

        previousRevision = blameModel.currentRevision
        saveAndTransposePosition &= previousRevision.isAnnotated()

        # Save current locator in nav history
        if saveAndTransposePosition:
            blameWindow.saveFilePosition()

        # Update scrubber
        # Heads up: Look up scrubber row from the sequence of nodes, not via
        # scrubber.findData(commitId, CommitLogModel.Role.Oid), because this
        # compares references to Oid objects, not Oid values.
        scrubberIndex = blameModel.revList.sequence.index(revision)
        with QSignalBlockerContext(blameWindow.scrubber):
            blameWindow.scrubber.setCurrentIndex(scrubberIndex)

        # Sync back/forward/newer/older states with scrubber and history BEFORE
        # loading the file to keep the user from spamming a button that's
        # supposed to be disabled.
        blameWindow.syncNavButtons()

        # Load the annotated revision if we haven't cached it before.
        # Note that the user can kill the task here, either by closing the
        # BlameWindow, or by switching to another commit using the scrubber
        # or nav buttons.
        if not revision.isAnnotated():
            blameWindow.busySpinner.start()
            yield from self._annotate(revision, blameModel)
            assert revision.isAnnotated()

        # OK, we haven't been interrupted and we're ready to display the file.
        # Make this TraceNode current in the model.
        blameModel.currentRevision = revision

        # Figure out which line number (QTextBlock) to scroll to
        topBlock = blameWindow.textEdit.topLeftCornerCursor().blockNumber()
        if saveAndTransposePosition:  # Attempt to restore position across files
            try:
                oldLine = previousRevision.blameLines[1 + topBlock]
                topBlock = revision.findLine(oldLine, topBlock) - 1
            except (IndexError,  # Zero lines in annotatedFile ("File deleted in commit" notice)
                    ValueError):  # Could not findLineByReference
                pass  # default to raw line number already stored in topBlock

        # Get file text
        useLexer = False
        if revision.binary:
            text = _("Binary blob")
            text = f"*** {text} ***"
        elif revision.fullText is None:
            if revision.commitId == NULL_OID:
                text = _("File deleted in working directory")
            else:
                text = _("File deleted in commit {0}", shortHash(revision.commitId))
            text = f"*** {text} ***"
        else:
            text = revision.fullText
            useLexer = bool(text)

        newLocator = blameModel.currentRevision.toLocator()
        newLocator = blameWindow.navHistory.refine(newLocator)
        blameWindow.navHistory.push(newLocator)  # this should update in place

        blameWindow.textEdit.setPlainText(text)
        blameWindow.textEdit.currentLocator = newLocator
        blameWindow.textEdit.restorePosition(newLocator)
        blameWindow.textEdit.syncViewportMarginsWithGutter()

        title = _("Blame {path} @ {commit} (Revision {rev}/{total})",
                  path=tquo(revision.path),
                  commit=shortHash(revision.commitId) if revision.commitId else _("Uncommitted"),
                  rev=blameModel.revList.revisionNumber(revision.commitId),
                  total=len(blameModel.revList))
        blameWindow.setWindowTitle(title)

        if saveAndTransposePosition:
            blockPosition = blameWindow.textEdit.document().findBlockByNumber(topBlock).position()
            blameWindow.textEdit.restoreScrollPosition(blockPosition)

        # Install lex job
        lexJob = self._getLexJob(revision) if useLexer else None
        if lexJob is not None:
            blameWindow.textEdit.highlighter.installLexJob(lexJob)
            blameWindow.textEdit.highlighter.rehighlight()

        blameWindow.busySpinner.stop()
        blameWindow.syncNavButtons()

        # Refresh search term now that the document has changed
        blameWindow.textEdit.searchBar.reevaluateSearchTerm()

    def _annotate(self, revision: Revision, blameModel: BlameModel):
        assert not revision.isAnnotated(), "node annotation already built"

        # Initialize list with a dummy line #0 so effective numbering can start at 1
        dummyLine0 = Revision.BlameLine(revision.commitId, 0)
        revision.blameLines.append(dummyLine0)

        if revision.status == GitStatus.Deleted:
            return

        driver = yield from self.flowCallGit(
            "blame",
            "--porcelain",
            *argsIf(revision.commitId != UC_FAKEID, str(revision.commitId)),
            "-S", blameModel.revsFile.fileName(),
            "--",
            revision.path)

        stdout = driver.stdoutScrollback()

        allLines = []
        binaryCheckChars = 8000  # similar to git's buffer_is_binary

        for hexHash, originalLineNumber, lineText in parseGitBlame(stdout):
            oid = Oid(hex=hexHash)
            annotatedLine = Revision.BlameLine(oid, originalLineNumber)
            revision.blameLines.append(annotatedLine)

            if revision.binary:
                continue

            allLines.append(lineText)

            if binaryCheckChars < 0:
                pass
            elif lineText.find("\0", 0, binaryCheckChars) >= 0:
                revision.binary = True
                allLines.clear()  # don't care about the text anymore
            else:
                binaryCheckChars -= len(lineText)

        if not revision.binary:
            revision.fullText = "".join(allLines)

    @staticmethod
    def _getLexJob(revision: Revision) -> LexJob | None:
        if not settings.prefs.isSyntaxHighlightingEnabled():
            return None

        assert revision.fullText
        cacheKey = f"blame:{revision.commitId}:{revision.path}"

        try:
            return LexJobCache.get(cacheKey)
        except KeyError:
            pass

        lexer = LexerCache.getLexerFromPath(revision.path, settings.prefs.pygmentsPlugins)
        if lexer is None:
            return None

        lexJob = LexJob(lexer, revision.fullText, cacheKey)
        if lexJob is None:
            return None

        LexJobCache.put(lexJob)
        return lexJob



# -----------------------------------------------------------------------------
# Copyright (C) 2025 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.qt import *

class QFaintSeparator(QFrame):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setFrameStyle(QFrame.Shape.HLine)
        self.setFrameShadow(QFrame.Shadow.Plain)
        self.setMaximumHeight(1)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import enum
import itertools
import logging
from collections.abc import Iterable, Iterator

from gitfourchette import settings
from gitfourchette.appconsts import *
from gitfourchette.gitdriver import GitDelta
from gitfourchette.graph import Graph, GraphSpliceLoop, MockCommit
from gitfourchette.graph.graphbuilder import CommitTraits
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.repoprefs import RepoPrefs
from gitfourchette.toolbox import *

logger = logging.getLogger(__name__)

UC_FAKEID = NULL_OID
"Fake Oid for Uncommitted Changes (aka Working Directory in the UI)."

UC_FAKEREF = "UC_FAKEREF"
"Fake reference name for Uncommitted Changes (aka Working Directory in the UI)."
# Actual refs are either "HEAD" or they start with "refs/",
# so this name should never collide with user refs.

BEGIN_SSH_SIGNATURE = b"-----BEGIN SSH SIGNATURE-----"


def toggleSetElement(s: set, element):
    assert isinstance(s, set)
    try:
        s.remove(element)
        return False
    except KeyError:
        s.add(element)
        return True


class GpgStatus(enum.IntEnum):
    Unsigned        = 0
    Pending         = enum.auto()
    CantCheck       = enum.auto()
    MissingKey      = enum.auto()
    GoodTrusted     = enum.auto()
    GoodUntrusted   = enum.auto()
    ExpiredSig      = enum.auto()
    ExpiredKey      = enum.auto()
    RevokedKey      = enum.auto()
    Bad             = enum.auto()
    ProcessError    = enum.auto()

    def iconName(self):
        return _GpgStatusIconTable.get(self, "gpg-verify-cantcheck")

    def iconHtml(self):
        return stockIconImgTag(self.iconName())


_GpgStatusIconTable = {
    GpgStatus.Pending       : "gpg-verify-pending",
    GpgStatus.GoodTrusted   : "gpg-verify-good-trusted",
    GpgStatus.GoodUntrusted : "gpg-verify-good-untrusted",
    GpgStatus.ExpiredSig    : "gpg-verify-expired",
    GpgStatus.ExpiredKey    : "gpg-verify-expired",
    GpgStatus.RevokedKey    : "gpg-verify-bad",
    GpgStatus.Bad           : "gpg-verify-bad",
    GpgStatus.ProcessError  : "gpg-verify-cantcheck",
}


class CommitPathspecFilter(QObject):
    resultsUpdated = Signal()

    needle: str
    matchingIds: set[Oid] | None
    filterOnly: bool

    def __init__(self, parent=None):
        super().__init__(parent)
        self.clear()

    def clear(self):
        self.needle = ""
        self.matchingIds = None
        self.filterOnly = False

    def wantFilter(self) -> bool:
        return bool(self.needle) and self.matchingIds is not None and self.filterOnly

    def isReady(self) -> bool:
        return bool(self.needle) and self.matchingIds is not None

    def isQueryPending(self) -> bool:
        return bool(self.needle) and self.matchingIds is None


class RepoModel:
    repo: Repo

    walker: Walker | None
    """Walker used to generate the graph. Call initializeWalker before use.
    Keep it around to speed up ulterior refreshes."""

    commitSequence: list[CommitTraits]
    "Ordered list of commits."

    truncatedHistory: bool

    graph: Graph

    refs: dict[str, Oid]
    "Get target commit ID by reference name."

    refsAt: dict[Oid, list[str]]
    "Get all reference names pointing at a given commit ID."

    mergeheads: list[Oid]

    stashes: list[Oid]

    submodules: dict[str, str]
    "Map of submodule names to the relative paths of their workdirs."

    initializedSubmodules: set[str]
    "Set of submodule names that are registered in .gitmodules."

    remotes: list[str]
    "List of remote names."

    upstreams: dict[str, str]
    "Table of local branch names to upstream shorthand names."

    aheadBehind: dict[str, tuple[int, int]]
    """Table of local branch names to the number of commits the branch is
    ahead/behind of its upstream."""

    superproject: str
    "Path of the superproject. Empty string if this isn't a submodule."

    foreignCommits: set[Oid]
    """Use this to look up which commits are part of local branches,
    and which commits are 'foreign'."""

    hiddenRefs: set[str]
    "All cached refs that are hidden, either explicitly or via ref patterns."

    hideSeeds: set[Oid]
    localSeeds: set[Oid]

    hiddenCommits: set[Oid]
    "All cached commit oids that are hidden."

    commitPathspecFilter: CommitPathspecFilter

    gpgStatusCache: dict[Oid, tuple[GpgStatus, str]]
    gpgVerifyQueue: set[Oid]

    workdirStale: bool
    "Flag indicating that the workdir should be refreshed before use."

    workdirStatusReady: bool
    "Flag indicating that stageDiff and dirtyDiff are available."

    workdirUnstagedDeltas: list[GitDelta]
    workdirStagedDeltas: list[GitDelta]

    headIsDetached: bool
    homeBranch: str

    prefs: RepoPrefs

    def __init__(self, repo: Repo):
        assert isinstance(repo, Repo)

        self.commitSequence = []
        self.truncatedHistory = True

        self.walker = None
        self.graph = Graph()

        self.headIsDetached = False
        self.homeBranch = ""

        self.superproject = ""

        self.workdirStale = True
        self.workdirStatusReady = False
        self.workdirNumChanges = -1
        self.workdirUnstagedDeltas = []
        self.workdirStagedDeltas = []

        self.refs = {}
        self.refsAt = {}
        self.mergeheads = []
        self.stashes = []
        self.submodules = {}
        self.initializedSubmodules = set()
        self.remotes = []
        self.upstreams = {}
        self.aheadBehind = {}  # filled in outside RepoModel (needs git call)

        self.hiddenRefs = set()
        self.hiddenCommits = set()
        self.hideSeeds = set()
        self.localSeeds = set()

        self.commitPathspecFilter = CommitPathspecFilter()

        self.gpgStatusCache = {}
        self.gpgVerifyQueue = set()

        self.repo = repo

        self.prefs = RepoPrefs.initForRepo(repo)

        if settings.prefs.refSortClearTimestamp > self.prefs.refSortClearTimestamp:
            self.prefs.clearRefSort()

        # Prime ref cache after loading prefs (prefs contain hidden ref patterns)
        self.syncRefs()
        self.syncMergeheads()
        self.syncStashes()
        self.syncSubmodules()
        self.syncRemotes()
        self.syncUpstreams()

    def __del__(self):
        self.commitPathspecFilter.deleteLater()
        logger.debug("__del__ RepoModel")

    @property
    def numUncommittedChanges(self) -> int:
        """ Number of unstaged+staged files. Negative means unknown count. """
        if self.workdirStatusReady:
            return self.workdirNumChanges
        else:
            return -1

    @property
    def numRealCommits(self):
        # The first item in the commit sequence is the "fake commit" for Uncommitted Changes.
        return max(0, len(self.commitSequence) - 1)

    @property
    def headCommitId(self) -> Oid:
        """ Oid of the currently checked-out commit. """
        return self.refs.get("HEAD", NULL_OID)

    @property
    def singleRemote(self) -> bool:
        return len(self.remotes) == 1

    @benchmark
    def syncRefs(self):
        """ Refresh cached refs (`refs` and `refsAt`).

        Return True if there were any changes in the refs since the last
        refresh, or False if nothing changed.
        """

        headWasDetached = self.headIsDetached
        self.headIsDetached = self.repo.head_is_detached

        if self.headIsDetached or self.repo.head_is_unborn:
            self.homeBranch = ""
        else:
            self.homeBranch = self.repo.head_branch_shorthand

        refs = self.repo.map_refs_to_ids(include_stashes=False)
        refs[UC_FAKEREF] = UC_FAKEID

        if refs == self.refs:
            # Make sure it's sorted in the exact same order...
            if APP_DEBUG:
                assert list(refs) == list(self.refs), "refs key order changed! how did that happen?"

            # Nothing to do!
            # Still, signal a change if HEAD just detached/reattached.
            return headWasDetached != self.headIsDetached

        # Build reverse ref cache.
        refsAt: dict[Oid, list[str]] = {}
        for k, v in refs.items():
            try:
                refsAt[v].append(k)
            except KeyError:
                # Not using defaultdict because we don't want to create a
                # million empty lists when the UI code consults this dict.
                refsAt[v] = [k]

        # Special case for HEAD: Make it appear first in reverse ref cache
        try:
            headId = refs["HEAD"]
        except KeyError:
            pass
        else:
            refsAt[headId].remove("HEAD")
            refsAt[headId].insert(0, "HEAD")

        # Store new cache
        self.refs = refs
        self.refsAt = refsAt

        # Since the refs have changed, we need to refresh hidden refs
        self.refreshHiddenRefCache()

        # Let caller know that the refs changed.
        return True

    @benchmark
    def syncMergeheads(self):
        mh = self.repo.listall_mergeheads()
        if mh != self.mergeheads:
            self.mergeheads = mh
            return True
        return False

    @benchmark
    def syncStashes(self):
        stashes = []
        for stash in self.repo.listall_stashes():
            stashes.append(stash.commit_id)
        if stashes != self.stashes:
            self.stashes = stashes
            return True
        return False

    @benchmark
    def syncSubmodules(self):
        submodules = self.repo.listall_submodules_dict()
        initializedSubmodules = {name for name, path in submodules.items() if self.repo.submodule_dotgit_present(path)}

        if submodules != self.submodules or initializedSubmodules != self.initializedSubmodules:
            self.submodules = submodules
            self.initializedSubmodules = initializedSubmodules
            return True

        return False

    @benchmark
    def syncRemotes(self):
        # We could infer remote names from refCache, but we don't want
        # to miss any "blank" remotes that don't have any branches yet.
        # RemoteCollection.names() is much faster than iterating on RemoteCollection itself
        remotes = self.repo.listall_remotes_fast()
        if remotes != self.remotes:
            self.remotes = remotes
            return True
        return False

    @benchmark
    def syncUpstreams(self):
        upstreams = self.repo.listall_upstreams_fast()
        if upstreams != self.upstreams:
            self.upstreams = upstreams
            return True
        return False

    @property
    def shortName(self) -> str:
        prefix = ""
        if self.superproject:
            superprojectNickname = settings.history.getRepoNickname(self.superproject)
            prefix = superprojectNickname + ": "

        return prefix + settings.history.getRepoNickname(self.repo.workdir)

    @benchmark
    def primeWalker(self) -> Walker:
        tipIds = self.refs.values()
        sorting = SortMode.TOPOLOGICAL

        if settings.prefs.chronologicalOrder:
            # In strictly chronological ordering, a commit may appear before its
            # children if it was "created" later than its children. The graph
            # generator produces garbage in this case. So, for chronological
            # ordering, keep TOPOLOGICAL in addition to TIME.
            sorting |= SortMode.TIME

        if self.walker is None:
            self.walker = self.repo.walk(None, sorting)
        else:
            self.walker.reset()
            self.walker.sort(sorting)  # this resets the walker IF ALREADY WALKING (i.e. next was called once)

        # In topological mode, the order in which the tips are pushed is
        # significant (last in, first out). The tips should be pre-sorted in
        # ASCENDING chronological order so that the latest modified branches
        # come out at the top of the graph in topological mode.
        for tip in tipIds:
            if tip == UC_FAKEID:
                continue
            self.walker.push(tip)

        return self.walker

    def uncommittedChangesMockCommit(self):
        try:
            head = self.refs["HEAD"]
            parents = [head] + self.mergeheads
        except KeyError:  # Unborn HEAD
            parents = []

        return MockCommit(UC_FAKEID, parents)

    @property
    def nextTruncationThreshold(self) -> int:
        n = self.numRealCommits * 2
        n -= n % -1000  # round up to next thousand
        return max(n, settings.prefs.maxCommits)

    def dangerouslyDetachedHead(self):
        if not self.headIsDetached:
            return False

        try:
            headTips = self.refsAt[self.headCommitId]
        except KeyError:
            return False

        if headTips != ["HEAD"]:
            return False

        try:
            frame = self.graph.getCommitFrame(self.headCommitId)
        except KeyError:
            # Head commit not in graph, cannot determine if dangerous, err on side of caution
            return True

        arcs = list(frame.arcsClosedByCommit())

        if len(arcs) == 0:
            return True

        if len(arcs) != 1:
            return False

        return arcs[0].openedBy == UC_FAKEID

    @benchmark
    def syncTopOfGraph(self, oldRefs: dict[str, Oid]) -> GraphSpliceLoop:
        # DO NOT call processEvents() here. While splicing a large amount of
        # commits, GraphView may try to repaint an incomplete graph.
        # GraphView somehow ignores setUpdatesEnabled(False) here!
        gsl = GraphSpliceLoop(self.graph, self.commitSequence,
                              oldHeads=oldRefs.values(), newHeads=self.refs.values(),
                              hideSeeds=self.getHiddenTips(), localSeeds=self.getLocalTips())
        coSplice = gsl.coSplice()
        coSplice.send(None)  # prime the generator

        try:
            coSplice.send(self.uncommittedChangesMockCommit())
        except StopIteration:
            # e.g. switching from a branch to Detached HEAD on the same commit as the branch
            pass
        else:
            walker = self.primeWalker()
            for commit in walker:
                try:
                    coSplice.send(commit)
                except StopIteration:
                    break
        coSplice.close()  # flush it

        self.commitSequence = gsl.commitSequence
        self.hideSeeds = gsl.hideSeeds
        self.localSeeds = gsl.localSeeds
        self.hiddenCommits = gsl.hiddenCommits
        self.foreignCommits = gsl.foreignCommits
        return gsl

    @benchmark
    def toggleHideRefPattern(self, refPattern: str, allButThis: bool = False):
        if not allButThis:
            self.prefs.showPatterns.clear()  # clear show patterns in non-allButThis mode
            toggleSetElement(self.prefs.hidePatterns, refPattern)
        else:
            wasShowingThis = refPattern in self.prefs.showPatterns
            self.prefs.hidePatterns.clear()  # clear hide patterns in allButThis mode
            self.prefs.showPatterns.clear()  # only retain a single show pattern
            if not wasShowingThis:
                toggleSetElement(self.prefs.showPatterns, refPattern)

        self.prefs.setDirty()
        self.refreshHiddenRefCache()

        # Sync hidden commits
        heads = self.refs.values()
        newHideSeeds = self.getHiddenTips()
        newLocalSeeds = self.getLocalTips()
        gsl = GraphSpliceLoop(self.graph, self.commitSequence, oldHeads=heads, newHeads=heads,
                              hideSeeds=newHideSeeds, localSeeds=newLocalSeeds)
        gsl.sendAll(self.commitSequence)  # send the same commit sequence
        self.hiddenCommits = gsl.hiddenCommits
        self.foreignCommits = gsl.foreignCommits
        self.hideSeeds = newHideSeeds
        self.localSeeds = newLocalSeeds

    @benchmark
    def refreshHiddenRefCache(self):
        showPatterns = self.prefs.showPatterns
        hidePatterns = self.prefs.hidePatterns
        numStalePatterns = 0

        assert type(self.hiddenRefs) is set
        self.hiddenRefs.clear()

        if showPatterns:
            matchingRefs, numStalePatterns = self._matchPatterns(showPatterns)
            self.hiddenRefs.update(self.refs.keys())
            self.hiddenRefs.difference_update(matchingRefs)
        elif hidePatterns:
            matchingRefs, numStalePatterns = self._matchPatterns(hidePatterns)
            self.hiddenRefs.update(matchingRefs)

        if numStalePatterns > 0:
            self.prefs.setDirty()

    def _matchPatterns(self, patterns):
        assert type(patterns) is set

        matchingRefs = set()
        patternsSeen = set()

        for ref in self.refs:
            # Test for verbatim match first
            if ref in patterns:
                matchingRefs.add(ref)
                patternsSeen.add(ref)
                continue

            # Break up directory segments
            i = len(ref)
            while i >= 0:
                i = ref.rfind('/', 0, i)
                if i < 0:
                    break
                prefix = ref[:i+1]
                if prefix in patterns:
                    matchingRefs.add(ref)
                    patternsSeen.add(prefix)
                    break

        numStalePatterns = len(patterns) - len(patternsSeen)
        assert numStalePatterns >= 0, "there can't be fewer patterns in total than we've visited!"
        if numStalePatterns != 0:
            logger.debug(f"Culling {numStalePatterns} stale ref patterns")
            patterns.intersection_update(patternsSeen)

        return matchingRefs, numStalePatterns

    def getKnownTips(self) -> Iterable[Oid]:
        return self.refs.values()

    def getLocalTips(self):
        return {
            oid for oid, refList in self.refsAt.items()
            if any(name == "HEAD" or name.startswith("refs/heads/") for name in refList)}

    def getHiddenTips(self) -> set[Oid]:
        seeds = set()
        hiddenRefs = self.hiddenRefs

        def isSharedByVisibleBranch(oid: Oid):
            return any(
                ref for ref in self.refsAt[oid]
                if ref not in hiddenRefs and not ref.startswith(RefPrefix.TAGS))

        for ref in hiddenRefs:
            oid = self.refs[ref]
            if not isSharedByVisibleBranch(oid):
                seeds.add(oid)

        return seeds

    def commitsMatchingRefPattern(self, refPattern: str) -> Iterator[Oid]:
        if not refPattern.endswith("/"):
            # Explicit ref
            try:
                yield self.refs[refPattern]
            except KeyError:
                pass
        else:
            # Wildcard
            for ref, oid in self.refs.items():
                if ref.startswith(refPattern):
                    yield oid

    def getCachedGpgStatus(self, commit: CommitTraits) -> tuple[GpgStatus, str]:
        oid = commit.id

        try:
            return self.gpgStatusCache[oid]
        except KeyError:
            pass

        sig, _payload = commit.gpg_signature
        status = GpgStatus.Pending if sig else GpgStatus.Unsigned

        if not sig:
            status = GpgStatus.Unsigned
            keyInfo = ""
        else:
            isSsh = sig.startswith(BEGIN_SSH_SIGNATURE)
            keyInfo = "ssh" if isSsh else ""

        self.gpgStatusCache[commit.id] = (status, keyInfo)
        return status, keyInfo

    def cacheGpgStatus(self, oid: Oid, status: GpgStatus, keyInfo: str = ""):
        self.gpgStatusCache[oid] = (status, keyInfo)

    def queueGpgVerification(self, oid: Oid):
        if not settings.prefs.verifyGpgOnTheFly:
            return
        assert self.gpgStatusCache[oid][0] == GpgStatus.Pending
        self.gpgVerifyQueue.add(oid)

    def findWorkdirDelta(self, path: str):
        """
        Return a delta for this path among the uncommitted changes.
        Return None if the file has no uncommitted changes.
        """

        staged = self.workdirStagedDeltas
        unstaged = self.workdirUnstagedDeltas

        # Look at UNSTAGED changes first as they are the freshest
        for deltas in (unstaged, staged):
            try:
                return next(d for d in deltas if path in (d.old.path, d.new.path))
            except StopIteration:
                pass

        # File has no uncommitted changes
        return None

    def workdirMatchesPathNeedle(self, needleLower: str) -> GitDelta | None:
        """
        Match a repo-relative path against the same pattern style as
        `git log -- <pathspec>`
        """

        assert needleLower == needleLower.lower()

        if not needleLower or not self.workdirStatusReady:
            return None

        for delta in itertools.chain(self.workdirUnstagedDeltas, self.workdirStagedDeltas):
            for df in delta.new, delta.old:
                if df.matchPathspec(needleLower):
                    return delta

        return None



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations


from gitfourchette import settings
from gitfourchette.codeview.codegutter import CodeGutter
from gitfourchette.qt import *

if TYPE_CHECKING:
    from gitfourchette.diffview.diffview import DiffView


class DiffGutter(CodeGutter):
    def __init__(self, parent):
        super().__init__(parent)
        self.maxLine = 0
        self.paddingString = "0 0"

    def refreshMetrics(self):
        lineNumber = self.maxLine
        if lineNumber == 0:
            maxDigits = 0
        else:
            maxDigits = len(str(lineNumber))
        self.paddingString = "0" * (2 * maxDigits + 2)

    def calcWidth(self) -> int:
        return self.fontMetrics().horizontalAdvance(self.paddingString)

    def paintEvent(self, event: QPaintEvent):
        diffView = self.codeView
        if TYPE_CHECKING:
            assert isinstance(diffView, DiffView)

        painter = QPainter(self)

        # Set up colors
        palette = self.palette()
        themeFG = palette.color(QPalette.ColorRole.Text)  # standard theme foreground color
        lineColor = QColor(themeFG.red(), themeFG.green(), themeFG.blue(), 80)
        textColor = QColor(themeFG.red(), themeFG.green(), themeFG.blue(), 80)

        # Gather some metrics
        rightEdge = self.rect().width() - 1
        fontHeight = self.fontMetrics().height()

        # Draw line numbers and hunk separator lines
        if settings.prefs.colorblind:
            noOldPlaceholder = "+"
            noNewPlaceholder = "-"
        else:
            noOldPlaceholder = "·"
            noNewPlaceholder = "·"

        linePen = QPen(lineColor, 1, Qt.PenStyle.DashLine)
        textPen = QPen(textColor)
        painter.setPen(textPen)

        for block, top, bottom in self.paintBlocks(event, painter, lineColor):
            blockNumber = block.blockNumber()

            if blockNumber >= len(diffView.lineData):
                break

            ld = diffView.lineData[blockNumber]
            if ld.origin:
                # Draw line numbers
                old = str(ld.oldLineNo) if ld.oldLineNo > 0 else noOldPlaceholder
                new = str(ld.newLineNo) if ld.newLineNo > 0 else noNewPlaceholder

                colW = (rightEdge - 3) // 2
                painter.drawText(0, top, colW, fontHeight, Qt.AlignmentFlag.AlignRight, old)
                painter.drawText(colW, top, colW, fontHeight, Qt.AlignmentFlag.AlignRight, new)
            else:
                # Draw hunk separator horizontal line
                y = round((top + bottom) / 2)
                painter.setPen(linePen)
                painter.drawLine(QLine(0, y, rightEdge, y))
                painter.setPen(textPen)  # restore text pen

        painter.end()



# -----------------------------------------------------------------------------
# Copyright (C) 2024 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

# Color scheme based on https://clrs.cc/

from gitfourchette.qt import QColor

navy    = QColor(0x001F3F)
blue    = QColor(0x0074D9)
aqua    = QColor(0x7FDBFF)
teal    = QColor(0x39CCCC)
olive   = QColor(0x3D9970)
green   = QColor(0x2ECC40)
lime    = QColor(0x00EE66)
yellow  = QColor(0xFFBB00)
orange  = QColor(0xFF851B)
red     = QColor(0xFF4136)
fuchsia = QColor(0xF012BE)
purple  = QColor(0xB10DC9)
maroon  = QColor(0x85144B)
white   = QColor(0xFFFFFF)
silver  = QColor(0xDDDDDD)
gray    = QColor(0xAAAAAA)
black   = QColor(0x111111)

rainbow = [
    navy, blue, aqua, teal, olive, green, lime, yellow, orange, red, maroon, fuchsia, purple
]

rainbowBright = [
    orange, yellow, lime, teal, blue, purple, fuchsia, red
]

grayscale = [
    black, gray, silver, white
]



# Form implementation generated from reading ui file 'commitinfodialog.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_CommitInfoDialog(object):
    def setupUi(self, CommitInfoDialog):
        CommitInfoDialog.setObjectName("CommitInfoDialog")
        CommitInfoDialog.resize(480, 342)
        CommitInfoDialog.setWindowTitle("CommitInfoDialog")
        self.gridLayout = QGridLayout(CommitInfoDialog)
        self.gridLayout.setHorizontalSpacing(20)
        self.gridLayout.setObjectName("gridLayout")
        self.iconLabel = QLabel(parent=CommitInfoDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.iconLabel.sizePolicy().hasHeightForWidth())
        self.iconLabel.setSizePolicy(sizePolicy)
        self.iconLabel.setText("ICON")
        self.iconLabel.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignTop)
        self.iconLabel.setObjectName("iconLabel")
        self.gridLayout.addWidget(self.iconLabel, 0, 0, 1, 1)
        self.summaryLabel = QLabel(parent=CommitInfoDialog)
        self.summaryLabel.setText("<html><head/><body><p>INFO<br/>INFO<br/>INFO<br/>INFO</p></body></html>")
        self.summaryLabel.setTextFormat(Qt.TextFormat.RichText)
        self.summaryLabel.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignTop)
        self.summaryLabel.setWordWrap(True)
        self.summaryLabel.setOpenExternalLinks(False)
        self.summaryLabel.setTextInteractionFlags(Qt.TextInteractionFlag.TextBrowserInteraction)
        self.summaryLabel.setObjectName("summaryLabel")
        self.gridLayout.addWidget(self.summaryLabel, 0, 1, 2, 1)
        self.buttonBox = QDialogButtonBox(parent=CommitInfoDialog)
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout.addWidget(self.buttonBox, 2, 0, 1, 2)
        self.detailsEdit = QTextEdit(parent=CommitInfoDialog)
        self.detailsEdit.setMinimumSize(QSize(0, 200))
        self.detailsEdit.setReadOnly(True)
        self.detailsEdit.setObjectName("detailsEdit")
        self.gridLayout.addWidget(self.detailsEdit, 3, 0, 1, 2)

        self.retranslateUi(CommitInfoDialog)
        self.buttonBox.accepted.connect(CommitInfoDialog.accept)
        self.buttonBox.rejected.connect(CommitInfoDialog.reject)

    def retranslateUi(self, CommitInfoDialog):
        pass



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from pygments.token import Token

from gitfourchette import colors
from gitfourchette.exttools.usercommand import UserCommand
from gitfourchette.qt import *
from gitfourchette.toolbox.textutils import qstringLength


class UserCommandSyntaxHighlighter(QSyntaxHighlighter):
    def __init__(self, parent):
        super().__init__(parent)

        from gitfourchette.syntax import LexerCache

        self.lexer = LexerCache.getLexerFromPath("GitFourchetteUserCommandsSyntaxHighlighter.bash", False)

        self.commentFormat = QTextCharFormat()
        self.commentFormat.setForeground(QColor(0x808080))

        self.titleFormat = QTextCharFormat(self.commentFormat)
        self.titleFormat.setFontWeight(QFont.Weight.Bold)
        self.titleFormat.setFontItalic(True)

        self.badTokenFormat = QTextCharFormat()
        self.badTokenFormat.setForeground(colors.red)

        self.goodTokenFormat = QTextCharFormat()
        self.goodTokenFormat.setForeground(colors.blue)
        self.goodTokenFormat.setFontWeight(QFont.Weight.Bold)

        self.acceleratorFormat = QTextCharFormat(self.titleFormat)
        self.acceleratorFormat.setFontUnderline(True)

    def highlightBlock(self, text: str):
        tokens = self.lexer.get_tokens(text)
        start = 0
        isCommandLine = False

        for tokenType, token in tokens:
            tokenLength = qstringLength(token)

            if token.startswith("$"):
                isValid = token in UserCommand.Token._value2member_map_
                self.setFormat(start, tokenLength,
                               self.goodTokenFormat if isValid else self.badTokenFormat)

            elif token.startswith(UserCommand.AlwaysConfirmPrefix) and not isCommandLine:
                self.setFormat(start, len(UserCommand.AlwaysConfirmPrefix), self.goodTokenFormat)

            elif tokenType == Token.Comment.Single:
                self.setFormat(start, tokenLength,
                               self.commentFormat if not isCommandLine else self.titleFormat)

                accelMatch = UserCommand.AcceleratorPattern.search(token)
                if accelMatch:
                    matchStart, matchEnd = accelMatch.span()
                    matchStart, matchEnd = qstringLength(token[:matchStart]), qstringLength(token[:matchEnd])
                    self.setFormat(start + matchStart, matchEnd - matchStart, self.acceleratorFormat)

            # Once we've seen at least one text token, we know that's a command line
            if tokenType == Token.Text:
                isCommandLine = True

            start += tokenLength



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.blameview.blamemodel import BlameModel
from gitfourchette.graphview.commitlogdelegate import CommitLogDelegate
from gitfourchette.graphview.graphpaint import paintGraphFrame
from gitfourchette.localization import *
from gitfourchette.porcelain import Oid
from gitfourchette.qt import *
from gitfourchette.toolbox import stockIcon

_emptyOidSet: set[Oid] = set()


class BlameScrubberDelegate(CommitLogDelegate):
    def __init__(self, blameModel: BlameModel, parent: QWidget):
        self.blameModel = blameModel
        self.singleItem = False
        super().__init__(repoModel=blameModel.repoModel, parent=parent)

    def prepareForDeletion(self):
        del self.blameModel
        super().prepareForDeletion()

    def isBold(self, oid: Oid) -> bool:
        return not self.singleItem and oid == self.blameModel.currentRevision.commitId

    def isDim(self, oid: Oid) -> bool:
        return False

    def paintPrivate(
            self,
            painter: QPainter,
            option: QStyleOptionViewItem,
            index: QModelIndex,
            rect: QRect,
            oid: Oid,
    ):
        assert oid is not None
        revision = self.blameModel.revList.revisionForCommit(oid)

        # Graph frame
        graphRect = QRect(rect)
        paintGraphFrame(painter, graphRect, oid, self.blameModel.graph, _emptyOidSet)
        rect.setLeft(graphRect.right())

        # Icon
        iconRect = QRect(rect)
        iconSize = min(16, iconRect.height())
        iconRect.setWidth(iconSize)
        icon = stockIcon(f"status_{revision.status.lower()}")
        icon.paint(painter, iconRect, Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter)
        rect.setLeft(iconRect.right() + 5)

    def uncommittedChangesMessage(self) -> str:
        return _("Uncommitted Changes in Working Directory")



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette import settings
from gitfourchette.forms.brandeddialog import convertToBrandedDialog
from gitfourchette.forms.ui_stashdialog import Ui_StashDialog
from gitfourchette.localization import *
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.toolbox import *

# In a stash, changes from the index and the worktree are combined.
# Use the order of the keys in this dictionary to find out which status
# will "win out" in a stash for any given file.
# The order of the keys is significant! Last key takes precedence.
# (NB: Dict key order is stable since Python 3.7)
statusPrecedence = {
    FileStatus.INDEX_MODIFIED: 'm',
    FileStatus.INDEX_RENAMED: 'r',
    FileStatus.INDEX_DELETED: 'd',
    FileStatus.INDEX_TYPECHANGE: 't',
    FileStatus.WT_MODIFIED: 'm',
    # INDEX_NEW takes precedence over WT_MODIFIED: if you stage a new file
    # and then modify it, it will appear as a new file in the stash.
    FileStatus.INDEX_NEW: 'a',
    # Other WT flags take precedence over INDEX flags.
    FileStatus.WT_NEW: 'a',
    FileStatus.WT_RENAMED: 'r',
    FileStatus.WT_DELETED: 'd',
    FileStatus.WT_TYPECHANGE: 't',
}


class StashDialog(QDialog):
    def __init__(
            self,
            repoStatus: dict[str, FileStatus],
            preTicked: list[str],
            parent: QWidget):
        super().__init__(parent)

        self.ui = Ui_StashDialog()
        self.ui.setupUi(self)
        self.ui.fileList.setVerticalScrollMode(settings.prefs.listViewScrollMode)
        self.ui.keepCheckBox.setToolTip("<p>" + self.ui.keepCheckBox.toolTip())
        self.ui.indexAndWtWarning.setVisible(False)
        self.ui.indexAndWtWarning.setText("\u26a0 " + self.ui.indexAndWtWarning.text())
        tweakWidgetFont(self.ui.fileList, 95)

        for label in (self.ui.willStashLabel, self.ui.indexAndWtWarning):
            tweakWidgetFont(label, 90)

        okButton = self.ui.buttonBox.button(QDialogButtonBox.StandardButton.Ok)

        buttonCaptions = [_("Stash && Reset Changes"), _("Stash && Keep Changes")]
        helpCaptions = [
            _("&Include these files in the stash, then reset them to their unmodified state:"),
            _("&Include these files in the stash:")]
        self.ui.keepCheckBox.clicked.connect(lambda keep: okButton.setText(buttonCaptions[keep]))
        self.ui.keepCheckBox.clicked.connect(lambda keep: self.ui.willStashLabel.setText(helpCaptions[keep]))

        self.ui.fileList.setUniformItemSizes(True)
        scrollTo = None
        for filePath, fileStatus in repoStatus.items():
            prefix = ""
            if (fileStatus & FileStatus_INDEX_MASK) and (fileStatus & FileStatus_WT_MASK):
                self.ui.indexAndWtWarning.setVisible(True)
                prefix = "\u26a0 "

            listItem = QListWidgetItem(prefix + filePath, self.ui.fileList)
            listItem.setData(Qt.ItemDataRole.UserRole, filePath)
            if not preTicked or filePath in preTicked:
                listItem.setCheckState(Qt.CheckState.Checked)
                if not scrollTo:
                    scrollTo = listItem
            else:
                listItem.setCheckState(Qt.CheckState.Unchecked)

            # Pick an icon that reflects(ish) the future status of the file in the stash
            strongestLetter = ''
            for flag, iconLetter in statusPrecedence.items():
                if fileStatus & flag:
                    strongestLetter = iconLetter
            if strongestLetter:
                icon = stockIcon(f"status_{strongestLetter}")
                listItem.setIcon(icon)

            self.ui.fileList.addItem(listItem)

        # Prime checkbox signal connections
        self.ui.keepCheckBox.click()
        self.ui.keepCheckBox.click()

        convertToBrandedDialog(self)
        self.ui.messageEdit.setFocus()

        # Make sure at least one ticked item is visible
        if scrollTo:
            self.ui.fileList.scrollToItem(scrollTo, hint=QAbstractItemView.ScrollHint.EnsureVisible)

    def tickedPaths(self) -> list[str]:
        paths = []

        for i in range(self.ui.fileList.count()):
            listItem = self.ui.fileList.item(i)
            if listItem.checkState() != Qt.CheckState.Unchecked:
                path = listItem.data(Qt.ItemDataRole.UserRole)
                paths.append(path)

        return paths



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.qt import *
from gitfourchette.sidebar.sidebarmodel import SidebarNode, SidebarItem, SidebarModel
from gitfourchette.toolbox.filterchangecontext import FilterChangeContext

_filterableItems = {
    SidebarItem.LocalBranch,
    SidebarItem.RemoteBranch,
    SidebarItem.Tag,
    SidebarItem.Stash,
    SidebarItem.Remote,
    SidebarItem.RefFolder,
    SidebarItem.Submodule,
    SidebarItem.DetachedHead,
    SidebarItem.UnbornHead,
}

_refItems = {
    SidebarItem.LocalBranch,
    SidebarItem.RemoteBranch,
    SidebarItem.Tag,
    SidebarItem.Stash,
}


class SidebarFilter(QSortFilterProxyModel):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._filterText = ""
        self.setRecursiveFilteringEnabled(True)
        self.setFilterCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)

    def setFilterText(self, text: str):
        with FilterChangeContext(self):
            self._filterText = text.lower()

    def filterAcceptsRow(self, sourceRow: int, sourceParent: QModelIndex) -> bool:
        if not self._filterText:
            return True

        index = self.sourceModel().index(sourceRow, 0, sourceParent)

        if not index.isValid():
            return True

        node = SidebarNode.fromIndex(index)
        item = node.kind

        if item not in _filterableItems:
            return True

        displayText = index.data(Qt.ItemDataRole.DisplayRole)
        if not displayText:
            return True

        # For branches and tags, check both display text and full ref path
        # This ensures that branches like "wip/leaf" match when searching for "wip"
        if item in _refItems:
            refName = index.data(SidebarModel.Role.Ref)
            if refName and self._filterText in refName.lower():
                return True

        return self._filterText in displayText.lower()



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import os
import re
import traceback


def shortenTracebackPath(line: str) -> str:
    return re.sub(r'^\s*File "([^"]+)", line (\d+)',
                  lambda m: F'{os.path.basename(m.group(1))}:{m.group(2)}',
                  line, count=1)


def excStrings(exc: Exception) -> tuple[str, str]:
    lines = traceback.format_exception_only(exc.__class__, exc)
    summary = ''.join(lines).strip()

    lines = traceback.format_exception(exc.__class__, exc, exc.__traceback__)
    details = ''.join(lines).strip()

    return summary, details



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette import settings
from gitfourchette.application import GFApplication
from gitfourchette.diffview.specialdiff import SpecialDiffError, ImageDelta
from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.toolbox import stockIcon, escape, DocumentLinks


class SpecialDiffView(QTextBrowser):
    linkActivated = Signal(QUrl)

    documentLinks: DocumentLinks | None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.documentLinks = None
        self.anchorClicked.connect(self.onAnchorClicked)
        GFApplication.instance().restyle.connect(self.refreshPrefs)
        GFApplication.instance().prefsChanged.connect(self.refreshPrefs)
        self.refreshPrefs()

    def refreshPrefs(self):
        scheme = settings.prefs.syntaxHighlightingScheme()
        styleSheet = scheme.basicQss(self)
        self.setStyleSheet(styleSheet)
        self.htmlHeader = "<html><style>a { font-weight: bold; }</style>" + settings.prefs.addDelColorsStyleTag()

    def onAnchorClicked(self, link: QUrl):
        if self.documentLinks is not None and self.documentLinks.processLink(link):
            return
        self.linkActivated.emit(link)

    def replaceDocument(self, newDocument: QTextDocument):
        self.documentLinks = None

        if self.document():
            self.document().deleteLater()

        self.setDocument(newDocument)
        self.clearHistory()

        self.setOpenLinks(False)

    def displaySpecialDiffError(self, err: SpecialDiffError):
        document = QTextDocument(self)
        document.setObjectName("DiffErrorDocument")

        icon = stockIcon(err.icon)
        pixmap: QPixmap = icon.pixmap(48, 48)
        document.addResource(QTextDocument.ResourceType.ImageResource, QUrl("icon"), pixmap)

        markup = (
            f"{self.htmlHeader}"
            "<table width='100%'>"
            "<tr>"
            f"<td width='{pixmap.width()}px'><img src='icon'/></td>"
            "<td width='100%' style='padding-left: 8px; padding-top: 8px;'>"
            f"<big>{err.message}</big>"
            f"<br/>{err.details}"
            "</td>"
            "</tr>"
            "</table>")

        if err.preformatted:
            markup += F"<pre>{escape(err.preformatted)}</pre>"

        markup += err.longform

        document.setHtml(markup)
        self.replaceDocument(document)

        assert self.documentLinks is None
        self.documentLinks = err.links

        # Let DocumentLinks callbacks invoke RepoTasks using this QObject chain
        err.taskInvoker = document

    def displayImageDelta(self, delta: ImageDelta, swap=False):
        if not swap:
            image = delta.new.image or delta.old.image
        else:
            image = delta.old.image or delta.new.image
        hasBothSides = delta.old.image and delta.new.image
        showLfsStatus = delta.old.deltaFile.lfs or delta.new.deltaFile.lfs

        green, red = settings.prefs.addDelColors()
        borderColor = red if image is delta.old.image else green

        links = DocumentLinks()
        swapLink = links.new(lambda: self.displayImageDelta(delta, swap=not swap))

        markup = self.htmlHeader + "<center><table>"

        for file in delta.old, delta.new:
            if file.image is None:
                continue

            tag = "add" if file is delta.new else "del"

            name = (_("New image") if file is delta.new else
                    _("Old image") if delta.new.image else
                    _("Deleted image"))

            # Show LFS info if either side is an LFS pointer:
            if showLfsStatus and not file.deltaFile.lfs.isTentative():
                lfsInfo = "LFS" if file.deltaFile.lfs else _("not LFS")
                name = f"{name}, {lfsInfo}"
            name += _(":")

            if hasBothSides and file.image is not image:
                c1 = f"<a href='{swapLink}'>{name}</a>"
            else:
                c1 = f"<b><{tag}>{name}</{tag}></b>"

            c2 = f"{file.image.width()} &times; {file.image.height()} {_('pixels')},"
            c3 = self.locale().formattedDataSize(file.size)

            if hasBothSides and file.image is image:
                c4 = f"<b><{tag}>&darr; {_('shown below')} &darr;</{tag}></b>"
            else:
                c4 = ""

            markup += (f"<tr><td>{c1} </td>"
                       f"<td style='text-align: right'>{c2} </td>"
                       f"<td style='text-align: right'>{c3} </td>"
                       f"<td style='text-align: center'> {c4}</td>"
                       f"</tr>")

        markup += (
            "</table>"
            f"<table style='border: 4px solid {borderColor.name()}; border-collapse: collapse;'>"
            "<tr><td><img src='image'/></tr></td>"
            "</table>"
            "</center>")

        document = QTextDocument(self)
        document.setObjectName("ImageDiffDocument")
        document.setHtml(markup)

        if image is not None:
            image.setDevicePixelRatio(self.devicePixelRatio())
            document.addResource(QTextDocument.ResourceType.ImageResource, QUrl("image"), image)

        self.replaceDocument(document)

        assert self.documentLinks is None
        self.documentLinks = links



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import logging
import shlex

from gitfourchette.forms.brandeddialog import convertToBrandedDialog
from gitfourchette.forms.ui_pushdialog import Ui_PushDialog
from gitfourchette.gitdriver import argsIf
from gitfourchette.localization import *
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.repomodel import RepoModel
from gitfourchette.toolbox import *

logger = logging.getLogger(__name__)


class PushDialog(QDialog):
    def onPickLocalBranch(self):
        upstreamShorthand = self.currentLocalBranchUpstreamShorthand
        upstreamIndex = self.getUpstreamIndex(upstreamShorthand)
        self.ui.remoteBranchEdit.setCurrentIndex(upstreamIndex)
        self.onPickRemoteBranch(upstreamIndex)
        self.updateTrackCheckBox()

        self.remoteBranchNameValidator.run()

    def onPickRemoteBranch(self, index: int):
        remoteName = self.currentRemoteName
        remoteTooltip = self.ui.remoteBranchEdit.itemData(index, Qt.ItemDataRole.ToolTipRole)

        if self.willPushToNewBranch:
            upstreamShorthand = self.currentLocalBranchUpstreamShorthand
            if upstreamShorthand and not self.currentLocalBranch.upstream:
                _remoteName, suggestedName = split_remote_branch_shorthand(upstreamShorthand)
            else:
                suggestedName = self.currentLocalBranchName

            reservedNames = self.reservedRemoteBranchNames[remoteName]
            suggestedName = withUniqueSuffix(suggestedName, reservedNames)

            self.ui.remoteNameLabel.setText(f"\u21AA {remoteName}/")
            self.ui.remoteNameLabel.setToolTip(remoteTooltip)
            self.ui.newRemoteBranchStackedWidget.setCurrentIndex(0)
            self.ui.newRemoteBranchNameEdit.setText(suggestedName)
            self.ui.newRemoteBranchNameEdit.setFocus(Qt.FocusReason.TabFocusReason)
        else:
            self.ui.newRemoteBranchStackedWidget.setCurrentIndex(1)
            self.ui.remoteBranchEdit.setFocus(Qt.FocusReason.TabFocusReason)

        self.ui.remoteBranchEdit.setToolTip(remoteTooltip)
        self.updateTrackCheckBox()

        self.remoteBranchNameValidator.run()

    def updateTrackCheckBox(self, resetCheckedState=True):
        localBranch = self.currentLocalBranch
        lbName = localBranch.shorthand
        rbName = self.currentRemoteBranchFullName

        lbUpstream = self.currentLocalBranchUpstreamShorthand
        hasUpstream = bool(lbUpstream)
        isTrackingHomeBranch = hasUpstream and lbUpstream == self.currentRemoteBranchFullName

        # Convert to quoted names
        lbName = hquoe(lbName)
        rbName = hquoe(rbName)
        lbUpstream = hquoe(lbUpstream)

        if hasUpstream and not localBranch.upstream:
            lbUpstream += " " + _("(missing on remote)")

        if not resetCheckedState:
            willTrack = self.ui.trackCheckBox.isChecked()
        else:
            willTrack = self.willPushToNewBranch or isTrackingHomeBranch

        if not hasUpstream and willTrack:
            text = _("{0} will track {1}.", lbName, rbName)
        elif not hasUpstream and not willTrack:
            text = _("{0} currently does not track any remote branch.", lbName)
        elif isTrackingHomeBranch:
            text = _("{0} already tracks remote branch {1}.", lbName, lbUpstream)
        elif willTrack:
            text = _("{0} will track {1} instead of {2}.", lbName, rbName, lbUpstream)
        else:
            text = _("{0} currently tracks {1}.", lbName, lbUpstream)

        self.ui.trackingLabel.setWordWrap(True)
        self.ui.trackingLabel.setText("<small>" + text + "</small>")
        self.ui.trackingLabel.setContentsMargins(20, 0, 0, 0)
        self.ui.trackingLabel.setEnabled(not isTrackingHomeBranch)
        self.ui.trackCheckBox.setEnabled(not isTrackingHomeBranch)
        self.setOkButtonText()

        if resetCheckedState:
            with QSignalBlockerContext(self.ui.trackCheckBox):
                self.ui.trackCheckBox.setChecked(willTrack)

    @property
    def repo(self) -> Repo:
        return self.repoModel.repo

    @property
    def currentLocalBranchName(self) -> str:
        return self.ui.localBranchEdit.currentData()

    @property
    def currentLocalBranch(self) -> Branch:
        return self.repo.branches.local[self.currentLocalBranchName]

    @property
    def currentLocalBranchUpstreamShorthand(self) -> str:
        localBranch = self.currentLocalBranch
        try:
            upstreamName = localBranch.upstream_name
        except KeyError:  # "config value 'branch.XXXX.remote' was not found"
            return ""
        return upstreamName.removeprefix(RefPrefix.REMOTES)

    @property
    def willForcePush(self) -> bool:
        return not self.willPushToNewBranch and self.ui.forcePushCheckBox.isChecked()

    @property
    def willPushToNewBranch(self) -> bool:
        data = self.ui.remoteBranchEdit.currentData()
        assert isinstance(data, str)
        return data.endswith("/")

    @property
    def currentRemoteName(self):
        data = self.ui.remoteBranchEdit.currentData()
        assert isinstance(data, str)
        remoteName, _branchName = split_remote_branch_shorthand(data)
        return remoteName

    @property
    def currentRemoteBranchName(self) -> str:
        if self.willPushToNewBranch:
            return self.ui.newRemoteBranchNameEdit.text()
        else:
            data = self.ui.remoteBranchEdit.currentData()
            assert isinstance(data, str)
            _remoteName, branchName = split_remote_branch_shorthand(data)
            return branchName

    @property
    def currentRemoteBranchFullName(self) -> str:
        return self.currentRemoteName + "/" + self.currentRemoteBranchName

    def refspec(self, withForcePrefix=True):
        prefix = "+" if withForcePrefix and self.willForcePush else ""
        lbn = self.currentLocalBranchName
        rbn = self.currentRemoteBranchName
        return f"{prefix}refs/heads/{lbn}:refs/heads/{rbn}"

    def getUpstreamIndex(self, upstream: str) -> int:
        comboBox = self.ui.remoteBranchEdit

        # Find the upstream as-is
        index = comboBox.findData(upstream)
        if index >= 0:
            return index

        # Fall back to "New branch" item for last used remote in this repo
        fallbackRemote = self.repoModel.prefs.getShadowUpstream(self.currentLocalBranchName)
        index = comboBox.findData(fallbackRemote)
        if index >= 0:
            return index

        # The local branch is bound to an upstream that's missing from our combobox.
        # Fall back to "New branch" item for the remote.
        if upstream:
            fallbackRemote, _dummy = split_remote_branch_shorthand(upstream)
            fallbackRemote += "/"
            index = comboBox.findData(fallbackRemote)
            if index >= 0:
                return index

        # Just find the first "New branch" item
        for index in range(comboBox.count()):
            data = comboBox.itemData(index)
            if data.endswith("/"):
                return index

        raise NotImplementedError(f"no suitable fallback index for upstream '{upstream}'")

    def fillRemoteComboBox(self):
        comboBox = self.ui.remoteBranchEdit
        currentLocalBranch = self.currentLocalBranch
        currentUpstream = currentLocalBranch.upstream.shorthand if currentLocalBranch.upstream else ""

        branchIcon = stockIcon("vcs-branch")
        newBranchIcon = stockIcon("SP_FileDialogNewFolder")
        boldFont = QFont(comboBox.font())
        boldFont.setBold(True)

        with QSignalBlockerContext(comboBox):
            comboBox.clear()

            for remoteName, branchNames in self.repo.listall_remote_branches(value_style="shorthand").items():
                if comboBox.count() != 0:
                    comboBox.insertSeparator(comboBox.count())

                remoteUrl = self.repo.remotes[remoteName].url
                for shorthand in sorted(branchNames, key=naturalSort):
                    isTracked = shorthand == currentUpstream
                    caption = shorthand
                    if isTracked:
                        caption += " " + _("[tracked]")
                    index = comboBox.count()
                    comboBox.addItem(branchIcon, caption, shorthand)
                    comboBox.setItemData(index, boldFont if isTracked else None, Qt.ItemDataRole.FontRole)
                    comboBox.setItemData(index, remoteUrl, Qt.ItemDataRole.ToolTipRole)

                newBranchPayload = remoteName + "/"
                newBranchCaption = _("New remote branch on {0}", lquo(remoteName))
                comboBox.addItem(newBranchIcon, newBranchCaption, newBranchPayload)
                comboBox.setItemData(comboBox.count()-1, remoteUrl, Qt.ItemDataRole.ToolTipRole)

    def __init__(self, repoModel: RepoModel, branch: Branch, parent: QWidget):
        super().__init__(parent)
        repo = repoModel.repo
        self.repoModel = repoModel
        self.reservedRemoteBranchNames = repo.listall_remote_branches()
        self.widgetsWereEnabled: list[bool] = []

        self.ui = Ui_PushDialog()
        self.ui.setupUi(self)
        self.ui.trackingLabel.setMinimumHeight(self.ui.trackingLabel.height())
        self.ui.trackingLabel.setMaximumHeight(self.ui.trackingLabel.height())
        self.ui.statusForm.connectAbortButton(self.cancelButton())

        self.startOperationButton: QPushButton = self.ui.buttonBox.button(QDialogButtonBox.StandardButton.Ok)
        self.startOperationButton.setText(_("&Push"))
        self.startOperationButton.setIcon(stockIcon("git-push"))

        lbComboBox = self.ui.localBranchEdit
        for lbName in sorted(repo.branches.local, key=naturalSort):
            lbComboBox.addItem(lbName, lbName)
        lbComboBox.setCurrentIndex(lbComboBox.findData(branch.shorthand))

        self.ui.localBranchEdit.activated.connect(self.fillRemoteComboBox)
        self.ui.localBranchEdit.activated.connect(self.onPickLocalBranch)
        self.ui.remoteBranchEdit.activated.connect(self.onPickRemoteBranch)
        self.ui.newRemoteBranchNameEdit.textEdited.connect(lambda text: self.updateTrackCheckBox(False))
        self.ui.trackCheckBox.toggled.connect(lambda: self.updateTrackCheckBox(False))

        self.remoteBranchNameValidator = ValidatorMultiplexer(self)
        self.remoteBranchNameValidator.setGatedWidgets(self.ui.buttonBox.button(QDialogButtonBox.StandardButton.Ok))
        self.remoteBranchNameValidator.connectInput(self.ui.newRemoteBranchNameEdit, self.validateCustomRemoteBranchName)
        # don't prime the validator!

        # Set up comboboxes (act as if we picked the current branch in localBranchEdit)
        self.fillRemoteComboBox()
        self.onPickLocalBranch()

        self.ui.forcePushCheckBox.clicked.connect(self.setOkButtonText)

        self.ui.forcePushCheckBox.setText(self.ui.forcePushCheckBox.text() + " " + _("(with lease)"))

        self.setOkButtonText()

        convertToBrandedDialog(self)

        self.setWindowModality(Qt.WindowModality.WindowModal)

        self.ui.forcePushCheckBox.toggled.connect(self.updateCommandPreview)
        self.ui.trackCheckBox.toggled.connect(self.updateCommandPreview)
        self.ui.localBranchEdit.activated.connect(self.updateCommandPreview)
        self.ui.remoteBranchEdit.activated.connect(self.updateCommandPreview)
        self.ui.newRemoteBranchNameEdit.textEdited.connect(self.updateCommandPreview)
        self.updateCommandPreview()

    def okButton(self) -> QPushButton:
        return self.ui.buttonBox.button(QDialogButtonBox.StandardButton.Ok)

    def cancelButton(self) -> QPushButton:
        return self.ui.buttonBox.button(QDialogButtonBox.StandardButton.Cancel)

    def setOkButtonText(self):
        icon = "git-push"
        tip = ""

        if self.willForcePush:
            text = _("Force &push")
            icon = "achtung"
            tip = _("Force push: Destructive action!")
        elif self.willPushToNewBranch:
            text = _("&Push new branch")
        else:
            text = _("&Push")

        okButton = self.okButton()
        okButton.setText(text)
        okButton.setIcon(stockIcon(icon))
        okButton.setToolTip(tip)

    def validateCustomRemoteBranchName(self, name: str):
        if not self.ui.newRemoteBranchNameEdit.isVisibleTo(self):
            return ""

        reservedNames = self.reservedRemoteBranchNames.get(self.currentRemoteName, [])

        return nameValidationMessage(name, reservedNames,
                                     _("This name is already taken by another branch on this remote."))

    def isBusy(self) -> bool:
        return len(self.widgetsWereEnabled) > 0

    def setBusy(self, busy: bool):
        widgets = [self.ui.remoteBranchEdit,
                   self.ui.localBranchEdit,
                   self.ui.newRemoteBranchNameEdit,
                   self.ui.forcePushCheckBox,
                   self.ui.trackCheckBox,
                   self.startOperationButton]

        if busy:
            # Remember which widgets were disabled already
            self.widgetsWereEnabled = [w.isEnabled() for w in widgets]
            for w in widgets:
                w.setEnabled(False)
            self.ui.statusForm.initProgress(_("Contacting remote host…"))
        else:
            for w, enableW in zip(widgets, self.widgetsWereEnabled, strict=True):
                w.setEnabled(enableW)
            self.widgetsWereEnabled.clear()

    def reject(self):
        if self.isBusy():
            self.ui.statusForm.requestAbort()
        else:
            super().reject()

    def saveShadowUpstream(self):
        branch = self.currentLocalBranch
        if branch.upstream:
            self.repoModel.prefs.setShadowUpstream(branch.branch_name, "")
        else:
            self.repoModel.prefs.setShadowUpstream(branch.branch_name, self.currentRemoteBranchFullName)

    def updateCommandPreview(self):
        tokens = self.buildCommand(pretty=True)
        command = "git " + shlex.join(tokens)

        pushCaption = stripAccelerators(_("&Push"))
        message = _("Click {0} to run:", tquo(pushCaption)) + "<p style='white-space: pre-wrap;'><small>" + escape(command)
        self.ui.statusForm.setBlurb(message)

    def buildCommand(self, pretty: bool = False):
        # To build the command, we rely on whether trackCheckBox is enabled.
        # This may be inaccurate when PushDialog is busy (all inputs disabled).
        assert not self.isBusy(), "buildCommand while PushDialog is busy!"

        refspec = self.refspec(withForcePrefix=False)  # no '+' prefix -- we control force via arguments to git
        resetTrackingReference = self.ui.trackCheckBox.isEnabled() and self.ui.trackCheckBox.isChecked()

        return [
            "push",
            *argsIf(not pretty, "--porcelain"),
            *argsIf(not pretty, "--progress"),
            *argsIf(self.willForcePush, "--force-with-lease"),
            *argsIf(resetTrackingReference, "--set-upstream"),
            self.currentRemoteName,
            refspec
        ]



# Form implementation generated from reading ui file 'checkoutcommitdialog.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_CheckoutCommitDialog(object):
    def setupUi(self, CheckoutCommitDialog):
        CheckoutCommitDialog.setObjectName("CheckoutCommitDialog")
        CheckoutCommitDialog.setWindowModality(Qt.WindowModality.NonModal)
        CheckoutCommitDialog.setEnabled(True)
        CheckoutCommitDialog.resize(581, 161)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(CheckoutCommitDialog.sizePolicy().hasHeightForWidth())
        CheckoutCommitDialog.setSizePolicy(sizePolicy)
        CheckoutCommitDialog.setSizeGripEnabled(False)
        CheckoutCommitDialog.setModal(True)
        self.verticalLayout = QVBoxLayout(CheckoutCommitDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.mainTitle = QLabel(parent=CheckoutCommitDialog)
        self.mainTitle.setObjectName("mainTitle")
        self.verticalLayout.addWidget(self.mainTitle)
        self.mainGroupBox = QGroupBox(parent=CheckoutCommitDialog)
        self.mainGroupBox.setObjectName("mainGroupBox")
        self.verticalLayout_3 = QVBoxLayout(self.mainGroupBox)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setSizeConstraint(QLayout.SizeConstraint.SetDefaultConstraint)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.switchRadioButton = QRadioButton(parent=self.mainGroupBox)
        self.switchRadioButton.setObjectName("switchRadioButton")
        self.horizontalLayout.addWidget(self.switchRadioButton)
        self.switchComboBox = QComboBox(parent=self.mainGroupBox)
        self.switchComboBox.setObjectName("switchComboBox")
        self.horizontalLayout.addWidget(self.switchComboBox)
        self.verticalLayout_3.addLayout(self.horizontalLayout)
        self.detachHeadRadioButton = QRadioButton(parent=self.mainGroupBox)
        self.detachHeadRadioButton.setObjectName("detachHeadRadioButton")
        self.verticalLayout_3.addWidget(self.detachHeadRadioButton)
        self.resetHeadRadioButton = QRadioButton(parent=self.mainGroupBox)
        self.resetHeadRadioButton.setObjectName("resetHeadRadioButton")
        self.verticalLayout_3.addWidget(self.resetHeadRadioButton)
        self.mergeRadioButton = QRadioButton(parent=self.mainGroupBox)
        self.mergeRadioButton.setObjectName("mergeRadioButton")
        self.verticalLayout_3.addWidget(self.mergeRadioButton)
        self.createBranchRadioButton = QRadioButton(parent=self.mainGroupBox)
        self.createBranchRadioButton.setObjectName("createBranchRadioButton")
        self.verticalLayout_3.addWidget(self.createBranchRadioButton)
        self.verticalLayout.addWidget(self.mainGroupBox)
        self.recurseSubmodulesSpacer = QFrame(parent=CheckoutCommitDialog)
        self.recurseSubmodulesSpacer.setFrameShape(QFrame.Shape.NoFrame)
        self.recurseSubmodulesSpacer.setLineWidth(0)
        self.recurseSubmodulesSpacer.setObjectName("recurseSubmodulesSpacer")
        self.verticalLayout_4 = QVBoxLayout(self.recurseSubmodulesSpacer)
        self.verticalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        spacerItem = QSpacerItem(20, 16, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.MinimumExpanding)
        self.verticalLayout_4.addItem(spacerItem)
        self.verticalLayout.addWidget(self.recurseSubmodulesSpacer)
        self.recurseSubmodulesTitle = QLabel(parent=CheckoutCommitDialog)
        self.recurseSubmodulesTitle.setObjectName("recurseSubmodulesTitle")
        self.verticalLayout.addWidget(self.recurseSubmodulesTitle)
        self.recurseSubmodulesGroupBox = QGroupBox(parent=CheckoutCommitDialog)
        self.recurseSubmodulesGroupBox.setObjectName("recurseSubmodulesGroupBox")
        self.verticalLayout_2 = QVBoxLayout(self.recurseSubmodulesGroupBox)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.recurseSubmodulesCheckBox = QCheckBox(parent=self.recurseSubmodulesGroupBox)
        self.recurseSubmodulesCheckBox.setChecked(True)
        self.recurseSubmodulesCheckBox.setObjectName("recurseSubmodulesCheckBox")
        self.verticalLayout_2.addWidget(self.recurseSubmodulesCheckBox)
        self.verticalLayout.addWidget(self.recurseSubmodulesGroupBox)
        self.buttonBox = QDialogButtonBox(parent=CheckoutCommitDialog)
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setCenterButtons(False)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(CheckoutCommitDialog)
        self.buttonBox.rejected.connect(CheckoutCommitDialog.reject)
        self.buttonBox.accepted.connect(CheckoutCommitDialog.accept)
        self.switchRadioButton.toggled['bool'].connect(self.switchComboBox.setEnabled)
        CheckoutCommitDialog.setTabOrder(self.switchRadioButton, self.switchComboBox)
        CheckoutCommitDialog.setTabOrder(self.switchComboBox, self.detachHeadRadioButton)
        CheckoutCommitDialog.setTabOrder(self.detachHeadRadioButton, self.resetHeadRadioButton)
        CheckoutCommitDialog.setTabOrder(self.resetHeadRadioButton, self.mergeRadioButton)
        CheckoutCommitDialog.setTabOrder(self.mergeRadioButton, self.createBranchRadioButton)
        CheckoutCommitDialog.setTabOrder(self.createBranchRadioButton, self.recurseSubmodulesCheckBox)

    def retranslateUi(self, CheckoutCommitDialog):
        CheckoutCommitDialog.setWindowTitle(_p("CheckoutCommitDialog", "Check out commit"))
        self.mainTitle.setText(_p("CheckoutCommitDialog", "How do you want to check out this commit?"))
        self.switchRadioButton.setText(_p("CheckoutCommitDialog", "Switch to &branch:"))
        self.switchComboBox.setToolTip(_p("CheckoutCommitDialog", "List of branches that point to this commit."))
        self.detachHeadRadioButton.setText(_p("CheckoutCommitDialog", "Enter &detached HEAD here"))
        self.resetHeadRadioButton.setText(_p("CheckoutCommitDialog", "&Reset HEAD here…"))
        self.mergeRadioButton.setText(_p("CheckoutCommitDialog", "&Merge into HEAD…"))
        self.createBranchRadioButton.setText(_p("CheckoutCommitDialog", "Start &new branch here…"))
        self.recurseSubmodulesTitle.setText(_p("CheckoutCommitDialog", "After the checkout:"))
        self.recurseSubmodulesCheckBox.setText(_p("CheckoutCommitDialog", "Update submodules recursively"))



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

from gitfourchette import settings, colors
from gitfourchette.blameview.blamemodel import BlameModel
from gitfourchette.codeview.codegutter import CodeGutter
from gitfourchette.localization import *
from gitfourchette.porcelain import Oid
from gitfourchette.qt import *
from gitfourchette.repomodel import UC_FAKEID
from gitfourchette.toolbox import *


class BlameGutter(CodeGutter):
    model: BlameModel

    def __init__(self, parent):
        super().__init__(parent)

        font = QFontDatabase.systemFont(QFontDatabase.SystemFont.SmallestReadableFont)
        setFontFeature(font, "tnum")  # Tabular numbers
        self.setFont(font)

        self.boldFont = QFont(font)
        self.boldFont.setBold(True)

        self.model = None

        self.installEventFilter(self)

        self.columnMetrics = []
        self.preferredWidth = 0
        self.lineHeight = 12

        self.lineColor = QColor()
        self.textColor = QColor()
        self.boldTextColor = QColor()
        self.freshColor = QColor()
        self.heatColor = QColor()
        self.unknownColor = QColor()

        self.refreshMetrics()

    def syncFont(self, codeFont: QFont):
        pointSize = codeFont.pointSizeF()
        codeFont = self.font()
        codeFont.setPointSizeF(pointSize)
        self.boldFont.setPointSizeF(pointSize)
        super().syncFont(codeFont)

    def refreshMetrics(self):
        fontMetrics = self.fontMetrics()

        maxLineNumber = self.codeView.blockCount()

        dateWidth = fontMetrics.horizontalAdvance("2000-00-00 ")
        authorWidth = fontMetrics.horizontalAdvance("M" * 8)
        lnWidth = fontMetrics.horizontalAdvance(" " + "0" * len(str(maxLineNumber)))

        self.columnMetrics = []
        x = 2
        for w in (dateWidth, authorWidth, lnWidth):
            self.columnMetrics.append((x, w))
            x += w
        x += 3
        self.preferredWidth = x

        self.lineHeight = max(fontMetrics.height(), self.codeView.fontMetrics().height())

        # Cache foreground colors
        standardForegroundColor = self.palette().color(QPalette.ColorRole.Text)
        fgRgb = standardForegroundColor.getRgb()[:3]
        self.lineColor = QColor(*fgRgb, 80)
        self.textColor = QColor(*fgRgb, 160)
        self.boldTextColor = QColor(*fgRgb, 210)

        # Cache background colors (make copies so alpha can be changed)
        self.heatColor = QColor(colors.orange)
        self.freshColor = QColor(colors.aqua)
        self.freshColor.setAlphaF(.6 if isDarkTheme(self.palette()) else .8)
        self.unknownColor = QColor(colors.fuchsia)
        self.unknownColor.setAlphaF(.4 if isDarkTheme(self.palette()) else .6)

    def calcWidth(self) -> int:
        return self.preferredWidth

    def eventFilter(self, watched, event: QEvent):
        if event.type() == QEvent.Type.ToolTip and isinstance(event, QHelpEvent):
            return self.doToolTip(event)
        return False

    def paintEvent(self, event: QPaintEvent):
        revision = self.model.currentRevision
        if revision is None:
            return

        painter = QPainter(self)
        bgColor = self.heatColor

        # Gather some metrics
        rightEdge = self.rect().width() - 1
        lh = self.lineHeight

        lc2 = QColor(self.lineColor)
        lc2.setAlphaF(lc2.alphaF()/2)
        linePen = QPen(lc2)
        textPen = QPen(self.textColor)
        boldTextPen = QPen(self.boldTextColor)
        painter.setPen(textPen)

        lastCaptionDrawnAtLine = -1
        hunkCommitId = None
        hunkStartLine = 1

        alignRight = Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter

        topCommitId = revision.commitId
        topRevisionNumber = self.model.revList.revisionNumber(topCommitId)

        for block, top, bottom in self.paintBlocks(event, painter, self.lineColor):
            lineNumber = 1 + block.blockNumber()
            try:
                annotatedLine = revision.blameLines[lineNumber]
                lineCommitId = annotatedLine.commitId
            except IndexError:
                break

            if lineCommitId != hunkCommitId:
                hunkCommitId = lineCommitId
                hunkStartLine = lineNumber
                isCurrent = lineCommitId == topCommitId
                painter.setFont(self.boldFont if isCurrent else self.font())
                painter.setPen(boldTextPen if isCurrent else textPen)

                try:
                    revisionNumber = self.model.revList.revisionNumber(lineCommitId)
                except LookupError:
                    revisionNumber = -1

                # Compute heat color
                if revisionNumber == topRevisionNumber:
                    bgColor = self.freshColor
                elif revisionNumber == -1:
                    bgColor = self.unknownColor
                else:
                    heat = revisionNumber / (topRevisionNumber - 1)
                    heat = heat ** 2  # ease in cubic
                    bgColor = self.heatColor
                    bgColor.setAlphaF(lerp(.0, .6, heat))

            # Fill heat rectangle
            heatTop = top if lastCaptionDrawnAtLine >= 0 else 0
            painter.fillRect(QRect(0, heatTop, rightEdge, bottom-heatTop), bgColor)

            # Draw line number
            lineNumL, lineNumW = self.columnMetrics[-1]
            painter.drawText(lineNumL, top, lineNumW, lh, alignRight, str(lineNumber))

            # Draw caption + separator line
            if lastCaptionDrawnAtLine < hunkStartLine:
                self.drawBlameCaption(lineCommitId, painter, top, lh)

                # Hunk separator line
                if lastCaptionDrawnAtLine > 0:
                    penBackup = painter.pen()
                    painter.setPen(linePen)
                    painter.drawLine(QLine(0, top, rightEdge-1, top))
                    painter.setPen(penBackup)  # restore text pen

                lastCaptionDrawnAtLine = lineNumber

        painter.end()

    def drawBlameCaption(self, commitId: Oid, painter: QPainter, top: int, lh: int):
        alignLeft = Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter
        dateL, dateW = self.columnMetrics[0]
        nameL, nameW = self.columnMetrics[1]

        if commitId == UC_FAKEID:
            dateText = self.locale().toString(QDateTime.currentDateTime(), "yyyy-MM-dd")
            nameText = _("(Uncommitted)")
        else:
            commit = self.model.repo.peel_commit(commitId)
            sig = commit.author
            dateText = signatureDateFormat(sig, "yyyy-MM-dd", localTime=True)
            nameText = abbreviatePerson(sig, AuthorDisplayStyle.LastName)

        # Date
        FittedText.draw(painter, QRect(dateL, top, dateW, lh), alignLeft, dateText, bypassSetting=True)

        # Author
        FittedText.draw(painter, QRect(nameL, top, nameW, lh), alignLeft, nameText)

    def doToolTip(self, event: QHelpEvent):
        assert isinstance(event, QHelpEvent)

        pos = event.globalPos()
        editLocalPos = self.codeView.mapFromGlobal(pos)
        textCursor = self.codeView.cursorForPosition(editLocalPos)
        lineNumber = 1 + textCursor.blockNumber()

        try:
            commitId = self.model.currentRevision.blameLines[lineNumber].commitId
        except LookupError:
            return False

        try:
            revision = self.model.revList.revisionForCommit(commitId)
            revisionNumber = self.model.revList.revisionNumber(commitId)
        except LookupError:
            revision = None
            revisionNumber = -1

        text = "<table style='white-space: pre'>"

        muted = mutedToolTipColorHex()
        colon = _(":")
        def newLine(heading, caption):
            return f"<tr><td style='color:{muted}; text-align: right;'>{heading}{colon} </td><td>{caption}</td>"

        isWorkdir = commitId == UC_FAKEID
        if isWorkdir:
            text += newLine(_("commit"), _("Not Committed Yet"))
        else:
            commit = self.model.repo.peel_commit(commitId)
            text += newLine(_("commit"), shortHash(commitId))
            text += newLine(_("author"), commit.author.name)
            text += newLine(_("date"), signatureDateFormat(commit.author, settings.prefs.shortTimeFormat, localTime=False))
        text += newLine(_("file name"), revision.path if revision else _("(not available)"))
        text += newLine(_("revision"), revisionNumber if revisionNumber >= 0 else _("(not available)"))
        text += "</table>"
        if not isWorkdir:
            text += "<p>" + escape(commit.message.rstrip()).replace("\n", "<br>") + "</p>"

        QToolTip.showText(event.globalPos(), text, self)
        event.accept()
        return True



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.tasks.repotask import RepoTask, RepoTaskRunner, TaskPrereqs, TaskEffects, TaskInvocation
from gitfourchette.tasks.repotask import RepoGoneError

from gitfourchette.tasks.blametasks import (
    OpenBlame,
)
from gitfourchette.tasks.branchtasks import (
    DeleteBranch,
    DeleteBranchFolder,
    EditUpstreamBranch,
    FastForwardBranch,
    MergeBranch,
    NewBranchFromCommit,
    NewBranchFromHead,
    NewBranchFromRef,
    RecallCommit,
    RenameBranch,
    RenameBranchFolder,
    ResetHead,
    SwitchBranch,
)
from gitfourchette.tasks.committasks import (
    AmendCommit,
    CheckoutCommit,
    CherrypickCommit,
    DeleteTag,
    NewCommit,
    NewTag,
    RevertCommit,
    SetUpGitIdentity,
)
from gitfourchette.tasks.exporttasks import (
    ExportCommitAsPatch,
    ExportPatchCollection,
    ExportStashAsPatch,
    ExportWorkdirAsPatch,
)
from gitfourchette.tasks.misctasks import (
    EditRepoSettings,
    GetCommitInfo,
    NewIgnorePattern,
    QueryCommitsTouchingPath,
    VerifyGpgSignature,
    VerifyGpgQueue,
)
from gitfourchette.tasks.jumptasks import (
    Jump,
    JumpBack,
    JumpForward,
    JumpToHEAD,
    JumpToUncommittedChanges,
    RefreshRepo,
)
from gitfourchette.tasks.loadtasks import (
    DownloadLfsObjects,
    LoadPatchInNewWindow,
)
from gitfourchette.tasks.nettasks import (
    AutoFetchRemotes,
    DeleteRemoteBranch,
    RenameRemoteBranch,
    FetchRemotes,
    FetchRemoteBranch,
    PullBranch,
    PushBranch,
    PushRefspecs,
    UpdateSubmodule,
    UpdateSubmodulesRecursive,
)
from gitfourchette.tasks.remotetasks import NewRemote, EditRemote, DeleteRemote

from gitfourchette.tasks.indextasks import (
    AbortMerge,
    AcceptMergeConflictResolution,
    ApplyPatch,
    ApplyPatchFile,
    ApplyPatchFileReverse,
    DiscardFiles,
    DiscardModeChanges,
    HardSolveConflicts,
    OpenMergeTool,
    ApplyPatchData,
    StageFiles,
    UnstageFiles,
    UnstageModeChanges,
    RestoreRevisionToWorkdir,
    SaveRevisionAs,
    OpenRevisionInEditor,
    OpenInDiffTool,
)

from gitfourchette.tasks.stashtasks import (
    ApplyStash,
    DropStash,
    NewStash,
)

from gitfourchette.tasks.submoduletasks import (
    AbsorbSubmodule,
    RegisterSubmodule,
    RemoveSubmodule,
)

from gitfourchette.tasks.taskbook import TaskBook



# Form implementation generated from reading ui file 'clonedialog.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_CloneDialog(object):
    def setupUi(self, CloneDialog):
        CloneDialog.setObjectName("CloneDialog")
        CloneDialog.setWindowModality(Qt.WindowModality.NonModal)
        CloneDialog.setEnabled(True)
        CloneDialog.resize(640, 340)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(CloneDialog.sizePolicy().hasHeightForWidth())
        CloneDialog.setSizePolicy(sizePolicy)
        CloneDialog.setSizeGripEnabled(False)
        CloneDialog.setModal(True)
        self.formLayout = QFormLayout(CloneDialog)
        self.formLayout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)
        self.formLayout.setObjectName("formLayout")
        self.urlLabel = QLabel(parent=CloneDialog)
        self.urlLabel.setObjectName("urlLabel")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.LabelRole, self.urlLabel)
        self.urlLayout = QHBoxLayout()
        self.urlLayout.setObjectName("urlLayout")
        self.urlEdit = QComboBox(parent=CloneDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.urlEdit.sizePolicy().hasHeightForWidth())
        self.urlEdit.setSizePolicy(sizePolicy)
        self.urlEdit.setEditable(True)
        self.urlEdit.setInsertPolicy(QComboBox.InsertPolicy.NoInsert)
        self.urlEdit.setObjectName("urlEdit")
        self.urlLayout.addWidget(self.urlEdit)
        self.protocolButton = ProtocolButton(parent=CloneDialog)
        self.protocolButton.setText("https")
        self.protocolButton.setObjectName("protocolButton")
        self.urlLayout.addWidget(self.protocolButton)
        self.formLayout.setLayout(0, QFormLayout.ItemRole.FieldRole, self.urlLayout)
        self.pathLabel = QLabel(parent=CloneDialog)
        self.pathLabel.setObjectName("pathLabel")
        self.formLayout.setWidget(1, QFormLayout.ItemRole.LabelRole, self.pathLabel)
        self.pathLayout = QHBoxLayout()
        self.pathLayout.setObjectName("pathLayout")
        self.pathEdit = QLineEdit(parent=CloneDialog)
        self.pathEdit.setObjectName("pathEdit")
        self.pathLayout.addWidget(self.pathEdit)
        self.browseButton = QToolButton(parent=CloneDialog)
        self.browseButton.setPopupMode(QToolButton.ToolButtonPopupMode.MenuButtonPopup)
        self.browseButton.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
        self.browseButton.setObjectName("browseButton")
        self.pathLayout.addWidget(self.browseButton)
        self.formLayout.setLayout(1, QFormLayout.ItemRole.FieldRole, self.pathLayout)
        self.optionsLabel = QLabel(parent=CloneDialog)
        self.optionsLabel.setObjectName("optionsLabel")
        self.formLayout.setWidget(2, QFormLayout.ItemRole.LabelRole, self.optionsLabel)
        self.recurseSubmodulesCheckBox = QCheckBox(parent=CloneDialog)
        self.recurseSubmodulesCheckBox.setChecked(True)
        self.recurseSubmodulesCheckBox.setObjectName("recurseSubmodulesCheckBox")
        self.formLayout.setWidget(2, QFormLayout.ItemRole.FieldRole, self.recurseSubmodulesCheckBox)
        self.shallowCloneLayout = QHBoxLayout()
        self.shallowCloneLayout.setObjectName("shallowCloneLayout")
        self.shallowCloneCheckBox = QCheckBox(parent=CloneDialog)
        self.shallowCloneCheckBox.setText("(SHALLOW CLONE)")
        self.shallowCloneCheckBox.setObjectName("shallowCloneCheckBox")
        self.shallowCloneLayout.addWidget(self.shallowCloneCheckBox)
        self.shallowCloneDepthSpinBox = QSpinBox(parent=CloneDialog)
        self.shallowCloneDepthSpinBox.setProperty("showGroupSeparator", True)
        self.shallowCloneDepthSpinBox.setPrefix("")
        self.shallowCloneDepthSpinBox.setMinimum(1)
        self.shallowCloneDepthSpinBox.setMaximum(999999)
        self.shallowCloneDepthSpinBox.setStepType(QAbstractSpinBox.StepType.AdaptiveDecimalStepType)
        self.shallowCloneDepthSpinBox.setObjectName("shallowCloneDepthSpinBox")
        self.shallowCloneLayout.addWidget(self.shallowCloneDepthSpinBox)
        self.shallowCloneSuffix = QLabel(parent=CloneDialog)
        self.shallowCloneSuffix.setText("(SUFFIX)")
        self.shallowCloneSuffix.setObjectName("shallowCloneSuffix")
        self.shallowCloneLayout.addWidget(self.shallowCloneSuffix)
        self.formLayout.setLayout(3, QFormLayout.ItemRole.FieldRole, self.shallowCloneLayout)
        self.keyFilePicker = KeyFilePickerCheckBox(parent=CloneDialog)
        self.keyFilePicker.setObjectName("keyFilePicker")
        self.formLayout.setWidget(4, QFormLayout.ItemRole.FieldRole, self.keyFilePicker)
        spacerItem = QSpacerItem(20, 16, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        self.formLayout.setItem(5, QFormLayout.ItemRole.FieldRole, spacerItem)
        self.statusGroupBox = QGroupBox(parent=CloneDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.statusGroupBox.sizePolicy().hasHeightForWidth())
        self.statusGroupBox.setSizePolicy(sizePolicy)
        self.statusGroupBox.setObjectName("statusGroupBox")
        self.verticalLayout = QVBoxLayout(self.statusGroupBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.statusForm = StatusForm(parent=self.statusGroupBox)
        self.statusForm.setObjectName("statusForm")
        self.verticalLayout.addWidget(self.statusForm)
        self.formLayout.setWidget(6, QFormLayout.ItemRole.SpanningRole, self.statusGroupBox)
        self.buttonBox = QDialogButtonBox(parent=CloneDialog)
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setCenterButtons(False)
        self.buttonBox.setObjectName("buttonBox")
        self.formLayout.setWidget(7, QFormLayout.ItemRole.FieldRole, self.buttonBox)
        self.urlLabel.setBuddy(self.urlEdit)
        self.pathLabel.setBuddy(self.pathEdit)
        self.optionsLabel.setBuddy(self.recurseSubmodulesCheckBox)
        self.shallowCloneSuffix.setBuddy(self.shallowCloneDepthSpinBox)

        self.retranslateUi(CloneDialog)
        self.buttonBox.rejected.connect(CloneDialog.reject)
        CloneDialog.setTabOrder(self.urlEdit, self.protocolButton)
        CloneDialog.setTabOrder(self.protocolButton, self.pathEdit)
        CloneDialog.setTabOrder(self.pathEdit, self.browseButton)
        CloneDialog.setTabOrder(self.browseButton, self.recurseSubmodulesCheckBox)
        CloneDialog.setTabOrder(self.recurseSubmodulesCheckBox, self.shallowCloneCheckBox)
        CloneDialog.setTabOrder(self.shallowCloneCheckBox, self.shallowCloneDepthSpinBox)
        CloneDialog.setTabOrder(self.shallowCloneDepthSpinBox, self.keyFilePicker)

    def retranslateUi(self, CloneDialog):
        CloneDialog.setWindowTitle(_p("CloneDialog", "Clone repository"))
        self.urlLabel.setText(_p("CloneDialog", "Remote &URL:"))
        self.pathLabel.setText(_p("CloneDialog", "Clone in&to:"))
        self.browseButton.setText(_p("CloneDialog", "&Browse…"))
        self.optionsLabel.setText(_p("CloneDialog", "Options:"))
        self.recurseSubmodulesCheckBox.setText(_p("CloneDialog", "Initialize submodules recursively (if any)"))
        self.shallowCloneCheckBox.setToolTip(_p("CloneDialog", "<p>Tick this to download just the latest commits, not the entire history of the repository."))
        self.keyFilePicker.setText(_p("CloneDialog", "Log in to SSH remotes with custom &key file"))
from gitfourchette.forms.keyfilepickercheckbox import KeyFilePickerCheckBox
from gitfourchette.forms.protocolbutton import ProtocolButton
from gitfourchette.forms.statusform import StatusForm



# Form implementation generated from reading ui file 'statusform.ui'
#
# Created by: PyQt6 UI code generator 6.10.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_StatusForm(object):
    def setupUi(self, StatusForm):
        StatusForm.setObjectName("StatusForm")
        StatusForm.resize(400, 129)
        StatusForm.setWindowTitle("")
        self.verticalLayout = QVBoxLayout(StatusForm)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.stackedWidget = QStackedWidget(parent=StatusForm)
        self.stackedWidget.setObjectName("stackedWidget")
        self.blurbPage = QWidget()
        self.blurbPage.setObjectName("blurbPage")
        self.verticalLayout_3 = QVBoxLayout(self.blurbPage)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.blurbScrollArea = QScrollArea(parent=self.blurbPage)
        self.blurbScrollArea.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.blurbScrollArea.setFrameShape(QFrame.Shape.NoFrame)
        self.blurbScrollArea.setWidgetResizable(True)
        self.blurbScrollArea.setObjectName("blurbScrollArea")
        self.blurbScrollAreaWidgetContents = QWidget()
        self.blurbScrollAreaWidgetContents.setGeometry(QRect(0, 0, 388, 117))
        self.blurbScrollAreaWidgetContents.setObjectName("blurbScrollAreaWidgetContents")
        self.verticalLayout_4 = QVBoxLayout(self.blurbScrollAreaWidgetContents)
        self.verticalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.blurbLabel = QLabel(parent=self.blurbScrollAreaWidgetContents)
        self.blurbLabel.setText("BLURB")
        self.blurbLabel.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignTop)
        self.blurbLabel.setWordWrap(True)
        self.blurbLabel.setObjectName("blurbLabel")
        self.verticalLayout_4.addWidget(self.blurbLabel)
        self.blurbScrollArea.setWidget(self.blurbScrollAreaWidgetContents)
        self.verticalLayout_3.addWidget(self.blurbScrollArea)
        self.stackedWidget.addWidget(self.blurbPage)
        self.progressPage = QWidget()
        self.progressPage.setObjectName("progressPage")
        self.verticalLayout_2 = QVBoxLayout(self.progressPage)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.titleLabel = QLabel(parent=self.progressPage)
        self.titleLabel.setText("TITLE")
        self.titleLabel.setObjectName("titleLabel")
        self.verticalLayout_2.addWidget(self.titleLabel)
        self.statusLabel = QElidedLabel(parent=self.progressPage)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Ignored, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.statusLabel.sizePolicy().hasHeightForWidth())
        self.statusLabel.setSizePolicy(sizePolicy)
        self.statusLabel.setText("STATUS")
        self.statusLabel.setObjectName("statusLabel")
        self.verticalLayout_2.addWidget(self.statusLabel)
        self.commandLabel = QElidedLabel(parent=self.progressPage)
        self.commandLabel.setEnabled(False)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Ignored, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.commandLabel.sizePolicy().hasHeightForWidth())
        self.commandLabel.setSizePolicy(sizePolicy)
        self.commandLabel.setText("COMMAND")
        self.commandLabel.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignTop)
        self.commandLabel.setObjectName("commandLabel")
        self.verticalLayout_2.addWidget(self.commandLabel)
        spacerItem = QSpacerItem(20, 90, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.verticalLayout_2.addItem(spacerItem)
        self.progressBar = QProgressBar(parent=self.progressPage)
        self.progressBar.setMaximum(0)
        self.progressBar.setProperty("value", 0)
        self.progressBar.setTextVisible(False)
        self.progressBar.setObjectName("progressBar")
        self.verticalLayout_2.addWidget(self.progressBar)
        self.stackedWidget.addWidget(self.progressPage)
        self.verticalLayout.addWidget(self.stackedWidget)

        self.retranslateUi(StatusForm)
        self.stackedWidget.setCurrentIndex(1)

    def retranslateUi(self, StatusForm):
        pass
from gitfourchette.toolbox.qelidedlabel import QElidedLabel



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import warnings
import weakref
from typing import ClassVar

from gitfourchette.qt import *


class QSignalBlockerContext:
    """
    Context manager wrapper around QSignalBlocker.
    """

    nestingLevels: ClassVar[dict[weakref.ReferenceType[QObject], int]] = {}  # Map object id to nesting depth
    concurrentBlockers: ClassVar[int] = 0

    def __init__(self, *objectsToBlock: QObject):
        self.objectsToBlock = [weakref.ref(o) for o in objectsToBlock]

    def __enter__(self):
        QSignalBlockerContext.concurrentBlockers += 1

        for ref in self.objectsToBlock:
            o = ref()
            assert o is not None
            self.nestingLevels[ref] = self.nestingLevels.get(ref, 0) + 1
            if self.nestingLevels[ref] == 1:
                # Block signals if we're the first QSignalBlockerContext to refer to this object
                if o.signalsBlocked():  # pragma: no cover
                    warnings.warn(f"QSignalBlockerContext: object signals already blocked! {o}")
                o.blockSignals(True)

    def __exit__(self, excType, excValue, excTraceback):
        for ref in self.objectsToBlock:
            o = ref()
            if o is None:
                del self.nestingLevels[ref]
            else:
                self.nestingLevels[ref] -= 1
                assert self.nestingLevels[ref] >= 0
                # Unblock signals if we were holding last remaining reference to this object
                if self.nestingLevels[ref] == 0:
                    o.blockSignals(False)
                    del self.nestingLevels[ref]

        QSignalBlockerContext.concurrentBlockers -= 1
        assert QSignalBlockerContext.concurrentBlockers >= 0
        assert QSignalBlockerContext.concurrentBlockers != 0 or not QSignalBlockerContext.concurrentBlockers



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

"""
Submodule management tasks.
"""

from contextlib import suppress
from pathlib import Path

from gitfourchette.forms.brandeddialog import convertToBrandedDialog
from gitfourchette.forms.registersubmoduledialog import RegisterSubmoduleDialog
from gitfourchette.localization import *
from gitfourchette.porcelain import *
from gitfourchette.tasks.repotask import AbortTask, RepoTask, TaskPrereqs, TaskEffects
from gitfourchette.toolbox import *


class RegisterSubmodule(RepoTask):
    def prereqs(self) -> TaskPrereqs:
        return TaskPrereqs.NoUnborn | TaskPrereqs.NoConflicts

    def flow(self, path: str):
        yield from self._flow(path, absorb=False)

    def _flow(self, path: str, absorb: bool):
        thisWD = Path(self.repo.workdir)
        thisName = thisWD.name
        subWD = Path(thisWD / path)
        subName = subWD.name
        subRemotes = {}

        preferredRemote = ""
        with RepoContext(thisWD / subWD) as subRepo:
            assert not subRepo.is_bare
            for remote in subRepo.remotes:
                subRemotes[remote.name] = remote.url
            with suppress(Exception):
                localBranch = subRepo.branches.local[subRepo.head_branch_shorthand]
                upstreamRemoteName = localBranch.upstream.remote_name
                preferredRemote = subRemotes[upstreamRemoteName]

        if not subRemotes:
            message = paragraphs(
                _("{0} has no remotes.", bquo(subName)),
                _("Please open {0} and add a remote to it before absorbing it as a submodule.", bquo(subName)))
            raise AbortTask(message)

        reservedNames = set(self.repo.listall_submodules_dict().keys())
        dlg = RegisterSubmoduleDialog(
            workdirPath=path,
            superprojectName=thisName,
            remotes=subRemotes,
            absorb=absorb,
            reservedNames=reservedNames,
            parent=self.parentWidget())

        if preferredRemote:
            i = dlg.ui.remoteComboBox.findData(preferredRemote)
            if i >= 0:
                dlg.ui.remoteComboBox.setCurrentIndex(i)

        subtitle = _("Settings will be saved in {0}", tquo(".gitmodules"))
        convertToBrandedDialog(dlg, subtitleText=subtitle)
        yield from self.flowDialog(dlg)

        remoteUrl = dlg.remoteUrl
        customName = dlg.customName
        dlg.deleteLater()

        self.epilog.effects |= TaskEffects.Workdir | TaskEffects.Refs  # we don't have TaskEffects.Submodules so .Refs is the next best thing

        innerWD = str(subWD.relative_to(thisWD))
        yield from self.flowCallGit("submodule", "add", "--force", "--name", customName, "--", remoteUrl, innerWD)

        if absorb:
            yield from self.flowCallGit("submodule", "absorbgitdirs", "--", innerWD)


class AbsorbSubmodule(RegisterSubmodule):
    def flow(self, path: str):
        yield from self._flow(path, absorb=True)


class RemoveSubmodule(RepoTask):
    def prereqs(self) -> TaskPrereqs:
        return TaskPrereqs.NoUnborn | TaskPrereqs.NoConflicts

    def flow(self, submoduleName: str):
        submodule = self.repo.submodules[submoduleName]
        path = submodule.path

        yield from self.flowConfirm(
            text=paragraphs(
                _("Really remove submodule {0}?", bquo(submoduleName)),
                _("The submodule will be removed from {0} and its working copy will be deleted.", hquo(".gitmodules")),
                _("Any changes in the submodule that haven’t been pushed will be lost."),
                _("This cannot be undone!")),
            buttonIcon="SP_DialogDiscardButton",
            verb="Remove")

        self.epilog.effects |= (
                TaskEffects.Workdir
                | TaskEffects.Refs  # no TaskEffects.Submodules, so .Refs is the next best thing
                | TaskEffects.Head  # also force refresh libgit2 index TODO: better naming
        )

        # 1. Unregister from .git/config & remove worktree
        yield from self.flowCallGit("submodule", "deinit", "--", path)

        # 2. Unregister from .gitmodules (& remove worktree again)
        yield from self.flowCallGit("rm", "--", path)

        self.epilog.status = _("Submodule {0} removed.", tquo(submoduleName))



# -----------------------------------------------------------------------------
# Copyright (C) 2025 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

"""
Font family/size picker composed of a QFontComboBox, a QSpinBox and a reset button.
"""

from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.toolbox.iconbank import stockIcon
from gitfourchette.toolbox.qsignalblockercontext import QSignalBlockerContext


class FontPicker(QWidget):
    assign = Signal(str, int)

    def __init__(self, parent):
        from gitfourchette.settings import qtIsNativeMacosStyle

        super().__init__(parent)

        self.defaultFont = QFontDatabase.systemFont(QFontDatabase.SystemFont.FixedFont)
        self.currentFont = QFont(self.defaultFont)

        self.familyEdit = FontPicker.FamilyComboBox(self)
        self.familyEdit.currentFontChanged.connect(self.updateFamily)
        self.familyEdit.setMaximumWidth(250)
        self.familyEdit.setMaxVisibleItems(25)
        self.familyEdit.setToolTip(_("Hold {key} to show proportional fonts.",
                                     key=QKeySequence("Shift").toString(QKeySequence.SequenceFormat.NativeText)))

        self.sizeEdit = QSpinBox(self)
        self.sizeEdit.setRange(6, 72)
        self.sizeEdit.valueChanged.connect(self.updateSize)
        if not qtIsNativeMacosStyle():
            self.sizeEdit.setMinimumHeight(self.familyEdit.height())
        self.sizeEdit.setSuffix(" " + _("pt"))

        self.resetButtonIconEnabled = stockIcon("SP_LineEditClearButton")
        self.resetButtonIconDisabled = QIcon()
        if qtIsNativeMacosStyle():
            # Mac style QToolButton ignores autoRaise, so never set an empty icon
            self.resetButtonIconDisabled = self.resetButtonIconEnabled

        self.resetButton = QToolButton(self)
        self.resetButton.setIcon(self.resetButtonIconEnabled)
        self.resetButton.setAutoRaise(True)
        self.resetButton.setToolTip(_("Reset to system default monospace font"))
        self.resetButton.clicked.connect(self.resetFont)

        layout = QHBoxLayout(self)
        layout.addWidget(self.familyEdit)
        layout.addWidget(self.sizeEdit)
        layout.addWidget(self.resetButton)
        layout.setContentsMargins(0, 0, 0, 0)
        self.refreshControls()

    def setCurrentFont(self, family: str, size: int) -> QFont:
        font = QFont(self.defaultFont)
        if family:
            font.setFamily(family)
        if size > 0:
            font.setPointSize(size)
        self.currentFont = font
        self.refreshControls()
        return font

    def resetFont(self):
        self.currentFont = QFont(self.defaultFont)
        self.refreshControls()
        self.fireAssign()

    def updateFamily(self, newFont: QFont):
        self.currentFont.setFamily(newFont.family())
        self.refreshControls(updateFontComboBox=False)
        self.fireAssign()

    def updateSize(self, newSize: int):
        self.currentFont.setPointSize(newSize)
        self.refreshControls()
        self.fireAssign()

    def refreshControls(self, updateFontComboBox=True):
        isDefaultFont = self.isDefault()

        self.resetButton.setEnabled(not isDefaultFont)
        self.resetButton.setIcon(self.resetButtonIconDisabled if isDefaultFont else self.resetButtonIconEnabled)

        font = self.currentFont
        with QSignalBlockerContext(self.sizeEdit):
            self.sizeEdit.setValue(font.pointSize())

        if updateFontComboBox:
            strippedFont = QFont(font)
            strippedFont.setPointSize(10)
            with QSignalBlockerContext(self.familyEdit):
                self.familyEdit.setCurrentFont(strippedFont)

    def isDefault(self):
        cf = self.currentFont
        df = self.defaultFont
        return cf.family() == df.family() and cf.pointSize() == df.pointSize()

    def fireAssign(self):
        if self.isDefault():
            self.assign.emit("", 0)
        else:
            cf = self.currentFont
            self.assign.emit(cf.family(), cf.pointSize())

    class FamilyComboBox(QFontComboBox):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.allFamilies = QFontDatabase.families()
            self.monospaceFamilies = [fam for fam in self.allFamilies if QFontDatabase.isFixedPitch(fam)]

        def showPopup(self):
            # QFontComboBox automagically resolves alias families like "Monospace" to a concrete font.
            # The concrete family is set as the current text, so preserve it (instead of currentFont.family()).
            currentFamily = self.currentText()

            showAll = QGuiApplication.keyboardModifiers() == Qt.KeyboardModifier.ShiftModifier

            families = self.allFamilies if showAll else self.monospaceFamilies

            with QSignalBlockerContext(self):
                self.clear()
                if not showAll and currentFamily not in families:
                    # Add current (likely proportional) font first
                    self.addItem(currentFamily)
                self.addItems(families)
                self.setCurrentText(currentFamily)

            super().showPopup()



# -----------------------------------------------------------------------------
# Copyright (C) 2024 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from pathlib import Path

from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.toolbox import QFilePickerCheckBox, PersistentFileDialog, paragraphs


class KeyFilePickerCheckBox(QFilePickerCheckBox):
    DefaultSshDir = "~/.ssh"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        filePickerTip = paragraphs(
            _("{app} normally uses public/private keys in {path} to authenticate you with remote servers.",
              app=qAppName(), path=KeyFilePickerCheckBox.DefaultSshDir),
            _("Tick this box if you want to access remotes in this repo with a custom key."))
        self.checkBox.setToolTip(filePickerTip)

    def fileDialog(self):
        sshDir = Path(KeyFilePickerCheckBox.DefaultSshDir).expanduser()
        fallbackPath = str(sshDir) if sshDir.exists() else ""

        prompt = _("Select public key file for this remote")
        publicKeyFilter = _("Public key file") + " (*.pub)"
        return PersistentFileDialog.openFile(self, "KeyFile", prompt, filter=publicKeyFilter, fallbackPath=fallbackPath)

    def validatePath(self, path: str):
        if not path:
            return ""

        p = Path(path)

        if not p.is_file():
            return _("File not found.")

        if path.endswith(".pub"):
            privateKey = p.with_suffix("")
            if not privateKey.is_file():
                return _("Accompanying private key not found.")
        else:
            privateKey = p
            if not privateKey.with_suffix(".pub").is_file():
                return _("Accompanying public key not found.")

        return ""

    def privateKeyPath(self):
        return self.path().removesuffix(".pub")



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import importlib
import logging
import sys


# For PyInstaller
def customStartModule():
    try:
        fullArg = sys.argv[1]
    except IndexError:
        return False

    magic = "--start-module="

    if not fullArg.startswith(magic):
        return False

    moduleName = fullArg.removeprefix(magic)
    if not moduleName.startswith("gitfourchette."):
        print("unsupported start module", file=sys.stderr)
        sys.exit(1)

    sys.argv.pop(1)
    sys.orig_argv.remove(fullArg)
    module = importlib.import_module(moduleName)
    module.main()
    return True


# Called from AppImage entry point
def main():
    logging.basicConfig(
        stream=sys.stderr,
        level=logging.WARNING,
        format='%(levelname).1s %(asctime)s %(filename)-16s | %(message)s',
        datefmt="%H:%M:%S")
    logging.captureWarnings(True)

    from gitfourchette.application import GFApplication
    GFApplication.main()


if __name__ == "__main__":
    if not customStartModule():
        main()



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

def main():
    import argparse
    import ctypes
    import sys

    parser = argparse.ArgumentParser(description="Send CTRL_C_EVENT or CTRL_BREAK_EVENT to win32 console process (roughly equivalent to SIGTERM)")
    parser.add_argument("pid", type=int, help="Process ID to terminate")
    parser.add_argument("-e", "--event", type=int, choices=[0, 1], default=0, help="0=CTRL_C_EVENT, 1=CTRL_BREAK_EVENT")
    args = parser.parse_args()

    def check(rc, exitCode):
        if rc == 0:
            sys.exit(exitCode)

    kernel = ctypes.windll.kernel32  # type: ignore[attr-defined]

    result = kernel.FreeConsole()
    check(result, 1)

    result = kernel.AttachConsole(args.pid)
    check(result, 2)

    result = kernel.SetConsoleCtrlHandler(None, 1)
    check(result, 3)

    result = kernel.GenerateConsoleCtrlEvent(args.event, 0)
    check(result, 4)


if __name__ == '__main__':
    main()
